# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import gettext
from json.tool import main
import os

from PyQt6.QtWidgets import (
    QButtonGroup,
    QFormLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QRadioButton,
    QScrollArea,
    QTabWidget,
    QVBoxLayout,
    QWidget,
    QWizard,
    QWizardPage,
)


from . import students
from . import widgets
from . import scores
from . import FileNameFilters
from .. import utils
from .. import exams

t = gettext.translation("eyegrade", utils.locale_dir(), fallback=True)
_ = t.gettext


class NewSessionPageInitial(QWizardPage):
    """First page of WizardNewSession.

    It asks for the directory in which the session has to be stored and
    the exam config file.

    """

    def __init__(self, config_filename=None):
        super().__init__()
        self.setTitle(_("Directory and exam configuration"))
        self.setSubTitle(
            _(
                "Select or create an empty directory in which you "
                "want to store the session "
                "and configure the exam from a file or manually."
            )
        )
        self.directory = widgets.OpenFileWidget(
            self,
            select_directory=True,
            title=_("Select or create an empty directory"),
            check_file_function=self._check_directory,
        )
        self.config_file = widgets.OpenFileWidget(
            self,
            title=_("Select the exam configuration file"),
            name_filter=FileNameFilters.exam_config,
            check_file_function=self._check_exam_config_file,
        )
        if config_filename is not None:
            self.config_file.set_text(config_filename)
        self.registerField("directory*", self.directory.filename_widget)
        if config_filename is None:
            self.registerField("config_file", self.config_file.filename_widget)
        self.combo = widgets.CustomComboBox(parent=self)
        self.combo.set_items(
            [
                _("Use an existing configuration file"),
                _("Configure a exam manually"),
                _("Configure a survey manually"),
            ]
        )
        self.combo.setCurrentIndex(0)
        self.combo.currentIndexChanged.connect(self._update_combo)
        self.config_file_label = QLabel(_("Exam configuration file"))
        layout = QFormLayout(self)
        self.setLayout(layout)
        layout.addRow(_("Directory"), self.directory)
        layout.addRow(self.combo)
        layout.addRow(self.config_file_label, self.config_file)
        self.registerField("exam_or_survey_combo", self.combo)

    def validatePage(self):
        """Called by QWizardPage to check the values of this page."""
        if not self.directory.is_validated():
            ok_dir = self.directory.check_value()
        else:
            ok_dir = True
        if self.combo.currentIndex() == 0:
            if not self.config_file.is_validated():
                ok_config = self.config_file.check_value()
            else:
                ok_config = True
            if ok_dir and ok_config:
                self.wizard().exam_config = exams.ExamConfig(
                    filename=self.config_file.text()
                )
                return True
            else:
                return False
        else:
            return ok_dir

    def _check_directory(self, dir_name):
        valid = True
        msg = ""
        if not os.path.isdir(dir_name):
            valid = False
            msg = _("The directory does not exist or is not a directory.")
        else:
            dir_content = os.listdir(dir_name)
            if dir_content:
                valid = False
                msg = _(
                    "The directory is not empty. "
                    "Choose another directory or create a new one."
                )
        return valid, msg

    def _check_exam_config_file(self, file_name):
        valid = True
        msg = ""
        if not os.path.exists(file_name):
            valid = False
            msg = _("The exam configuration file does not exist.")
        elif not os.path.isfile(file_name):
            valid = False
            msg = _("The exam configuration file is not a regular file.")
        else:
            try:
                exams.ExamConfig(filename=file_name)
            except IOError:
                valid = False
                msg = _("The exam configuration file cannot be read.")
            except Exception as e:
                valid = False
                msg = _("The exam configuration file contains errors")
                if str(e):
                    msg += ":<br><br>" + str(e)
                else:
                    msg += "."
        self.wizard().exam_config_reset()
        return valid, msg

    def _update_combo(self, new_index):
        """Enable or disable the .eye file input"""
        if new_index == 0:
            self.config_file.setEnabled(True)
            self.config_file_label.setEnabled(True)
        else:
            self.config_file.setEnabled(False)
            self.config_file_label.setEnabled(False)
            self.config_file.set_text("")

    def nextId(self):
        if self.combo.currentIndex() == 0:
            return WizardNewSession.PageStudents
        else:
            return WizardNewSession.PageExamParams


class NewSessionPageStudents(QWizardPage):
    """Page for importing the student list."""

    def __init__(self):
        super().__init__()
        self.setTitle(_("Student list"))
        self.setSubTitle(
            _(
                "You can import your student list "
                "from Excel XLSX files and CSV files. "
                "Students can optionally be placed "
                "into separate student groups."
            )
        )
        layout = QVBoxLayout(self)
        self.setLayout(layout)
        tabs = students.StudentGroupsTabs(self)
        layout.addWidget(tabs)
        self.student_listings = tabs.student_listings

    def validatePage(self):
        """Called by QWizardPage to check the values of this page."""
        return True


class NewSessionPageExamParams(QWizardPage):
    """Wizard's page that ask for the params of the test"""

    def __init__(self):
        super().__init__()
        self.setTitle(_("Configuration of exam params"))
        self.setSubTitle(_("Enter the configuration parameters of the exam"))
        layout = QFormLayout(self)
        self.setLayout(layout)
        self.paramNEIDs = widgets.InputInteger(
            initial_value=0, min_value=0, max_value=16
        )
        self.registerField("paramNEIDs", self.paramNEIDs)
        self.paramNAlts = widgets.InputInteger(initial_value=3)
        self.registerField("paramNAlts", self.paramNAlts)
        self.paramNCols = widgets.InputCustomPattern(
            fixed_size=250,
            regex=r"[1-9][0-9]?(\,[1-9][0-9]?)+",
            placeholder=_("num,num,..."),
        )
        self.registerField("paramNCols", self.paramNCols)
        self.paramNPerm = widgets.InputInteger(min_value=1, initial_value=2)
        self.registerField("paramNPerm", self.paramNPerm)
        ## self.paramTPerm = widgets.CustomComboBox(parent=self)
        ## self.paramTPerm.set_items([
        ##     _('No (recommended)'),
        ##     _('Yes (experimental)'),
        ## ])
        ## self.paramTPerm.setCurrentIndex(0)
        ## self.registerField("paramTPerm", self.paramTPerm)
        layout.addRow(
            _("Number of digits of the student ID (0 to not scan IDs)"), self.paramNEIDs
        )
        layout.addRow(_("Number of choices per question"), self.paramNAlts)
        layout.addRow(_("Number of questions per answer box"), self.paramNCols)
        layout.addRow(_("Number of models of the exam"), self.paramNPerm)

    def initializePage(self):
        if self.field("exam_or_survey_combo") == 2:
            # It's a survey instead of an exam:
            self.paramNPerm.setValue(1)
            self.paramNPerm.setEnabled(False)

    def validatePage(self):
        if not self.paramNEIDs.text():
            QMessageBox.critical(
                self,
                _("Error in form"),
                _("The number of digits of the student id is empty"),
            )
            return False
        if not self.paramNAlts.text():
            QMessageBox.critical(
                self,
                _("Error in form"),
                _("The number of choices per question is empty"),
            )
            return False
        if not self.paramNCols.text():
            QMessageBox.critical(
                self,
                _("Error in form"),
                _(
                    "The number of questions per answer box is empty."
                    " Enter a comma-separated list of natural numbers."
                ),
            )
            return False
        if self.paramNCols.text().endswith(","):
            self.paramNCols.setText(self.paramNCols.text()[:-1])
        if not self.paramNPerm.text() and self.field("exam_or_survey_combo") != 2:
            # It's an exam but the number of models hasn't been set
            QMessageBox.critical(
                self, _("Error in form"), _("The number of exam models is empty")
            )
            return False
        # Create now the exam configuration object
        exam_config = exams.ExamConfig()
        exam_config.scores_mode = exams.ExamConfig.SCORES_MODE_NONE
        # Dimensions descriptor:
        dimensions = []
        for num_questions in self.paramNCols.text().split(","):
            dimensions.append("{},{}".format(self.paramNAlts.text(), num_questions))
        exam_config.set_dimensions(";".join(dimensions))
        # Student id length:
        exam_config.id_num_digits = self.field("paramNEIDs")
        # Survey mode:
        if self.field("exam_or_survey_combo") == 2:
            exam_config.survey_mode = True
        self.wizard().exam_config = exam_config
        return True

    def nextId(self):
        if self.field("exam_or_survey_combo") != 2:
            # It's an exam: enter solutions now
            return WizardNewSession.PageExamAnswers
        # It's a survey:
        return WizardNewSession.PageStudents


class NewSessionPageExamAnswers(QWizardPage):
    def __init__(self):
        super().__init__()
        self.setTitle(_("Selection of correct answers"))
        self.setSubTitle(_("Select the correct answers for each exam model"))
        layout = QFormLayout()
        self.setLayout(layout)
        self.tabs = QTabWidget()
        layout.addRow(self.tabs)
        self.paramNAlts = None
        self.paramNCols = None
        self.paramNPerm = None

    def initializePage(self):
        new_paramNAlts = self.field("paramNAlts")
        new_paramNCols = self.field("paramNCols")
        new_paramNPerm = self.field("paramNPerm")
        if (
            new_paramNAlts != self.paramNAlts
            or new_paramNCols != self.paramNCols
            or new_paramNPerm != self.paramNPerm
        ):
            self.paramNAlts = new_paramNAlts
            self.paramNCols = new_paramNCols
            self.paramNPerm = new_paramNPerm
            self._initialize()

    def _initialize(self):
        ## self.paramTPerm = self.field("paramTPerm")
        self.tabs.clear()
        self.total_answers = 0
        self.radioGroups = {}
        filas = self.paramNPerm
        for x in range(filas):
            mygroupbox = QScrollArea()
            mygroupbox.setWidget(QWidget())
            mygroupbox.setWidgetResizable(True)
            myform = QHBoxLayout(mygroupbox.widget())
            cols = self.paramNCols.split(",")
            ansID = 0
            radioGroupList = {}
            for col in cols:
                mygroupboxCol = QGroupBox()
                myformCol = QFormLayout()
                mygroupboxCol.setLayout(myformCol)
                for __ in range(int(col)):
                    ansID += 1
                    radioGroupList[ansID] = QButtonGroup()
                    layoutRow = QHBoxLayout()
                    for j in range(self.paramNAlts):
                        myradio = QRadioButton(chr(97 + j).upper())
                        layoutRow.addWidget(myradio)
                        radioGroupList[ansID].addButton(myradio)
                    self.total_answers += 1
                    myformCol.addRow(str(ansID), layoutRow)
                myform.addWidget(mygroupboxCol)
            self.radioGroups[chr(97 + x).upper()] = radioGroupList
            self.tabs.addTab(mygroupbox, _("Model ") + chr(97 + x).upper())

    def _get_values(self):
        response = dict()
        for k, v in self.radioGroups.items():
            answers = []
            for _, av in v.items():
                answer = abs(int(av.checkedId())) - 1
                if answer > 0:
                    answers.append({answer})
                else:
                    answers.append({})
            response[k] = answers
        return response

    def _check_count_answers(self):
        local_radio_groups = self._get_values()
        local_total_answers = 0
        for model_answers in local_radio_groups.values():
            for question_answer in model_answers:
                if len(question_answer) > 0:
                    local_total_answers += 1
        return self.total_answers == local_total_answers

    def validatePage(self):
        if not self._check_count_answers():
            valid = False
            QMessageBox.critical(
                self,
                _("Error"),
                _("You haven't entered the correct answer for some questions."),
            )
        else:
            valid = True
            # Add solutions to the exam's configuration:
            current_solutions = self._get_values()
            for model, solutions in current_solutions.items():
                self.wizard().exam_config.set_solutions(model, solutions)
        return valid

    def nextId(self):
        return WizardNewSession.PageStudents


class NewSessionPageScores(QWizardPage):
    """Page of WizardNewSession that asks for the scores for answers."""

    def __init__(self):
        super().__init__()
        self.setTitle(_("Scores for correct and incorrect answers"))
        self.setSubTitle(
            _(
                "Enter the scores of correct and incorrect "
                "answers. The program will compute scores based "
                "on them. Setting these scores is optional."
            )
        )
        self.scores_widget = None

    def initializePage(self):
        """Loads the values from the exam config, if any."""
        self.scores_widget = scores.EditScoresWidget(
            self.wizard().exam_config, parent=self
        )
        main_layout = QVBoxLayout(self)
        main_layout.addWidget(self.scores_widget)

        self.scores_widget.initialize()

    def validatePage(self):
        """Called by QWizardPage to check the values of this page.

        Checks the values and consolidates them if valid.

        """
        return self.scores_widget.validate_and_consolidate()

    def clear_base_scores(self):
        if self.scores_widget is not None:
            self.scores_widget.clear_base_scores()

    def nextId(self):
        return -1


class WizardNewSession(QWizard):
    """Wizard for the creation of a new session.

    It asks for a directory in which to store the session, the .eye file,
    student lists and scores for correct/incorrect answers.

    An initial value for the path of the .eye file can be passed as
    `config_filename`.

    """

    NUM_PAGES = 5
    (PageInitial, PageExamParams, PageExamAnswers, PageStudents, PageScores) = range(
        NUM_PAGES
    )

    def __init__(self, parent, config_filename=None):
        super().__init__(parent)
        self.exam_config = None
        self.setWindowTitle(_("Create a new session"))
        self.page_initial = self._create_page_initial(config_filename=config_filename)
        self.page_students = self._create_page_students()
        self.page_exam_params = self._create_page_exam_file()
        self.page_exam_answers = self._create_page_exam_answers()
        self.page_scores = self._create_page_scores()

        self.setPage(self.PageInitial, self.page_initial)
        self.setPage(self.PageExamParams, self.page_exam_params)
        self.setPage(self.PageExamAnswers, self.page_exam_answers)
        self.setPage(self.PageStudents, self.page_students)
        self.setPage(self.PageScores, self.page_scores)

        self.setStartId(self.PageInitial)

    def get_directory(self):
        return self.page_initial.directory.text()

    def get_config_file_path(self):
        return self.page_initial.config_file.text()

    def student_listings(self):
        return self.page_students.student_listings

    def values(self):
        values = {}
        values["directory"] = self.get_directory()
        values["config_file_path"] = self.get_config_file_path()
        values["config"] = self.exam_config
        values["student_listings"] = self.student_listings()
        return values

    def exam_config_reset(self):
        """Called when the exam config file is set or its value is changed."""
        self.page_scores.clear_base_scores()

    def _create_page_initial(self, config_filename=None):
        """Creates the first page, which asks for directory and .eye file."""
        return NewSessionPageInitial(config_filename=config_filename)

    def _create_page_exam_file(self):
        return NewSessionPageExamParams()

    def _create_page_exam_answers(self):
        return NewSessionPageExamAnswers()

    def _create_page_students(self):
        return NewSessionPageStudents()

    def _create_page_scores(self):
        """Creates the scores page, which asks for (in)correct scores."""
        return NewSessionPageScores()



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2022 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import fractions
from typing import Union, Optional


class PointsWords:
    singular_word: str
    plural_word: str

    def __init__(self, word_1: str, word_2: Optional[str] = None):
        """Set the singular and plural words for "point".

        `word_1` is the singular form and `word_2` is the plural form.
        If only the first one is provided, `word_1` will be used for both.

        """
        self.singular_word = word_1
        if word_2 is not None:
            self.plural_word = word_2
        else:
            self.plural_word = word_1

    def for_score(self, score: Union[float, fractions.Fraction]):
        if score == 1:
            return self.singular_word
        else:
            return self.plural_word



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#

import unittest
import os.path
import tempfile
import copy

import eyegrade.students as students
import eyegrade.utils as utils


class TestStudent(unittest.TestCase):
    def test_full_name(self):
        student = students.Student("0000", "John Doe", None, None, "john@example.com")
        self.assertEqual(student.name, "John Doe")
        self.assertEqual(student.id_and_name, "0000 John Doe")
        self.assertEqual(student.name_or_id, "John Doe")

    def test_first_and_last_name(self):
        student = students.Student("0000", None, "John", "Doe", "john@example.com")
        self.assertEqual(student.name, "John Doe")
        self.assertEqual(student.id_and_name, "0000 John Doe")
        self.assertEqual(student.name_or_id, "John Doe")
        self.assertEqual(student.last_comma_first_name, "Doe, John")

    def test_last_name(self):
        student = students.Student("0000", None, None, "Doe", "doe@example.com")
        self.assertEqual(student.name, "Doe")
        self.assertEqual(student.id_and_name, "0000 Doe")
        self.assertEqual(student.name_or_id, "Doe")
        self.assertEqual(student.last_comma_first_name, "Doe")

    def test_first_name(self):
        student = students.Student("0000", None, "John", "", "john@example.com")
        self.assertEqual(student.name, "John")
        self.assertEqual(student.id_and_name, "0000 John")
        self.assertEqual(student.name_or_id, "John")
        self.assertEqual(student.last_comma_first_name, "John")

    def test_no_name(self):
        student = students.Student("0000", None, None, None, "doe@example.com")
        self.assertEqual(student.name, "")
        self.assertEqual(student.id_and_name, "0000")
        self.assertEqual(student.name_or_id, "0000")
        self.assertEqual(student.last_comma_first_name, "")

    def test_name_errors(self):
        self.assertRaises(
            ValueError,
            students.Student,
            "0000",
            "John Doe",
            "John",
            "",
            "doe@example.com",
        )
        self.assertRaises(
            ValueError,
            students.Student,
            "0000",
            "John Doe",
            "",
            "Doe",
            "doe@example.com",
        )
        self.assertRaises(
            ValueError,
            students.Student,
            "0000",
            "John Doe",
            "John",
            "Doe",
            "doe@example.com",
        )


class TestReadStudentsFromFile(unittest.TestCase):
    def setUp(self):
        self.students = [
            ("101010101", "Donald Duck", "Donald", "Duck", "duck@d.com"),
            ("202020202", "Marty McFly", "Marty", "McFly", "fly@d.com"),
            ("313131313", "Peter Pan", "Peter", "Pan", "pan@pan.com"),
        ]
        self.bad_students = [("1010a0101", "Bad Boy", "Bad", "Boy", "boy@bad.com")]
        self.non_ascii_students = [
            ("404040404", "Rey León", "Rey", "León", "lion@jungle.com")
        ]

    def _get_test_file_path(self, filename):
        dirname = os.path.dirname(os.path.abspath(__file__))
        return os.path.join(dirname, filename)

    def test_empty(self):
        with tempfile.TemporaryDirectory() as dir_name:
            test_file = os.path.join(dir_name, "test_file")
            with open(test_file, mode="w") as f:
                f.write("")
            student_list = students.read_students(test_file)
        self.assertEqual(student_list, [])

    def test_empty_with_header(self):
        with tempfile.TemporaryDirectory() as dir_name:
            test_file = os.path.join(dir_name, "test_file")
            with open(test_file, mode="w") as f:
                f.write("header\n")
            student_list = students.read_students(test_file)
        self.assertEqual(student_list, [])

    def test_read_id(self):
        with tempfile.TemporaryDirectory() as dir_name:
            test_file = os.path.join(dir_name, "test_file")
            with open(test_file, mode="w") as f:
                f.write("\n".join([s[0] for s in self.students]))
            student_list = students.read_students(test_file)
        data = [_student_tuple(s) for s in student_list]
        key = [(s[0], "", "", "", "") for s in self.students]
        self.assertEqual(data, key)

    def test_read_id_full_name(self):
        with tempfile.TemporaryDirectory() as dir_name:
            test_file = os.path.join(dir_name, "test_file")
            with open(test_file, mode="w") as f:
                f.write("\n".join(["\t".join((s[0], s[1])) for s in self.students]))
            student_list = students.read_students(test_file)
        data = [_student_tuple(s) for s in student_list]
        key = [(s[0], s[1], "", "", "") for s in self.students]
        self.assertEqual(data, key)

    def test_read_id_full_name_email(self):
        with tempfile.TemporaryDirectory() as dir_name:
            test_file = os.path.join(dir_name, "test_file")
            with open(test_file, mode="w") as f:
                f.write(
                    "\n".join(["\t".join((s[0], s[1], s[4])) for s in self.students])
                )
            student_list = students.read_students(test_file)
        data = [_student_tuple(s) for s in student_list]
        key = [(s[0], s[1], "", "", s[4]) for s in self.students]
        self.assertEqual(data, key)

    def test_read_id_name_surname(self):
        with tempfile.TemporaryDirectory() as dir_name:
            test_file = os.path.join(dir_name, "test_file")
            with open(test_file, mode="w") as f:
                f.write(
                    "\n".join(["\t".join((s[0], s[2], s[3])) for s in self.students])
                )
            student_list = students.read_students(test_file)
        data = [_student_tuple(s) for s in student_list]
        key = [(s[0], "", s[2], s[3], "") for s in self.students]
        self.assertEqual(data, key)

    def test_read_id_name_surname_email(self):
        with tempfile.TemporaryDirectory() as dir_name:
            test_file = os.path.join(dir_name, "test_file")
            with open(test_file, mode="w") as f:
                f.write(
                    "\n".join(
                        ["\t".join((s[0], s[2], s[3], s[4])) for s in self.students]
                    )
                )
            student_list = students.read_students(test_file)
        data = [_student_tuple(s) for s in student_list]
        key = [(s[0], "", s[2], s[3], s[4]) for s in self.students]
        self.assertEqual(data, key)

    def test_read_with_comma_separator(self):
        with tempfile.TemporaryDirectory() as dir_name:
            test_file = os.path.join(dir_name, "test_file")
            with open(test_file, mode="w") as f:
                f.write(
                    "\n".join(
                        [",".join((s[0], s[2], s[3], s[4])) for s in self.students]
                    )
                )
            student_list = students.read_students(test_file)
        data = [_student_tuple(s) for s in student_list]
        key = [(s[0], "", s[2], s[3], s[4]) for s in self.students]
        self.assertEqual(data, key)

    def test_read_with_header(self):
        with tempfile.TemporaryDirectory() as dir_name:
            test_file = os.path.join(dir_name, "test_file")
            with open(test_file, mode="w") as f:
                f.write("id\tname\tsurname\temail\n")
                f.write(
                    "\n".join(
                        ["\t".join((s[0], s[2], s[3], s[4])) for s in self.students]
                    )
                )
            student_list = students.read_students(test_file)
        data = [_student_tuple(s) for s in student_list]
        key = [(s[0], "", s[2], s[3], s[4]) for s in self.students]
        self.assertEqual(data, key)

    def test_read_with_header_comma_name(self):
        with tempfile.TemporaryDirectory() as dir_name:
            test_file = os.path.join(dir_name, "test_file")
            with open(test_file, mode="w") as f:
                f.write("id\tname\temail\n")
                f.write(
                    "\n".join(
                        [
                            "\t".join((s[0], "{}, {}".format(s[3], s[2]), s[4]))
                            for s in self.students
                        ]
                    )
                )
            student_list = students.read_students(test_file)
        data = [_student_tuple(s) for s in student_list]
        key = [(s[0], "{}, {}".format(s[3], s[2]), "", "", s[4]) for s in self.students]
        self.assertEqual(data, key)

    def test_read_without_header_comma_name(self):
        with tempfile.TemporaryDirectory() as dir_name:
            test_file = os.path.join(dir_name, "test_file")
            with open(test_file, mode="w") as f:
                f.write(
                    "\n".join(
                        [
                            "\t".join((s[0], "{}, {}".format(s[3], s[2]), s[4]))
                            for s in self.students
                        ]
                    )
                )
            student_list = students.read_students(test_file)
        data = [_student_tuple(s) for s in student_list]
        key = [(s[0], "{}, {}".format(s[3], s[2]), "", "", s[4]) for s in self.students]
        self.assertEqual(data, key)

    def test_errors(self):
        with tempfile.TemporaryDirectory() as dir_name:
            test_file = os.path.join(dir_name, "test_file")
            with open(test_file, mode="w") as f:
                f.write("error\nerror\n")
            test_file_2 = os.path.join(dir_name, "test_file_2")
            with open(test_file_2, mode="w") as f:
                f.write(
                    "\n".join(
                        [
                            "\t".join((s[0], s[1], s[4]))
                            for s in self.students + self.bad_students
                        ]
                    )
                )
            self.assertRaises(
                utils.EyegradeException, students.read_students, test_file
            )
            self.assertRaises(
                utils.EyegradeException, students.read_students, test_file_2
            )

    def test_non_ascii(self):
        with tempfile.TemporaryDirectory() as dir_name:
            test_file = os.path.join(dir_name, "test_file")
            with open(test_file, mode="w") as f:
                f.write(
                    "\n".join(
                        [
                            "\t".join((s[0], s[2], s[3], s[4]))
                            for s in self.students + self.non_ascii_students
                        ]
                    )
                )
            student_list = students.read_students(test_file)
        data = [_student_tuple(s) for s in student_list]
        key = [
            (s[0], "", s[2], s[3], s[4])
            for s in self.students + self.non_ascii_students
        ]
        self.assertEqual(data, key)

    def test_xlsx(self):
        test_file = self._get_test_file_path("student-list.xlsx")
        student_list = students.read_students(test_file)
        key = [
            students.Student("100000333", "", "Frodo", "Baggins", ""),
            students.Student("100777777", "", "Bugs", "Bunny", ""),
            students.Student("100999997", "", "Bastian B.", "Bux", ""),
            students.Student("100999991", "", "Harry", "Potter", ""),
            students.Student("100800003", "", "Lisa", "Simpson", ""),
        ]
        self.assertEqual(student_list, key)


class TestListings(unittest.TestCase):
    def setUp(self):
        self.students = [
            students.Student("101010101", "Donald Duck", "", "", ""),
            students.Student("202020202", "Marty McFly", "", "", ""),
            students.Student("313131313", "Peter Pan", "", "", ""),
        ]
        self.more_students = [
            students.Student("909090909", "Donkey Kong", "", "", ""),
            students.Student("818181818", "Pinoccio", "", "", ""),
            students.Student("555555555", "Gump", "", "", ""),
        ]

    def test_listing_add(self):
        listing = students.GroupListing(students.StudentGroup(1, "G"), [])
        listing.add_students(self.students)
        self.assertEqual(listing.students, self.students)
        listing.add_students(self.more_students)
        self.assertEqual(listing.students, self.students + self.more_students)
        self.assertTrue(self.more_students[1].student_id in listing)

    def test_listing_find_duplicates(self):
        listing = students.GroupListing(None, list(self.students))
        new_students = [
            students.Student("313131313", "Peter", "", "", ""),
            self.more_students[0],
            self.more_students[1],
            students.Student("202020202", "Marty", "", "", ""),
        ]
        duplicates = listing.find_duplicates(new_students)
        key = [new_students[0], new_students[3]]
        self.assertEqual(duplicates, key)

    def test_listing_find_duplicates_2(self):
        listing = students.GroupListing(None, list(self.students))
        dup = copy.copy(self.more_students[0])
        dup.full_name = "Other Donkey"
        new_students = [
            self.more_students[0],
            self.more_students[1],
            dup,
            self.more_students[2],
        ]
        duplicates = listing.find_duplicates(new_students)
        key = [dup]
        self.assertEqual(duplicates, key)

    def test_listing_find_duplicates_3(self):
        listing = students.GroupListing(None, list(self.students))
        dup = copy.copy(self.students[1])
        dup.full_name = "Other Marty"
        new_students = [
            self.more_students[0],
            self.students[1],
            dup,
            self.more_students[1],
        ]
        duplicates = listing.find_duplicates(new_students)
        self.assertTrue(dup in duplicates)
        self.assertTrue(self.students[1] in duplicates)
        self.assertEqual(len(duplicates), 2)

    def test_listing_add_duplicates(self):
        listing = students.GroupListing(None, list(self.students))
        new_students = [
            students.Student("313131313", "Peter", "", "", ""),
            students.Student("909090909", "Donkey Kong", "", "", ""),
            students.Student("202020202", "Marty", "", "", ""),
        ]
        self.assertRaises(
            students.DuplicateStudentIdException, listing.add_students, new_students
        )

    def test_rank_students(self):
        # Test for issue #132
        listing = students.GroupListing(
            students.StudentGroup(1, "G"), list(self.students)
        )
        student_listings = students.StudentListings()
        student_listings.add_listing(listing)
        rank = zip([1.5, 1.5, 3], student_listings.iter_students())
        [student for score, student in sorted(rank, reverse=True)]


def _student_tuple(student):
    return (
        student.student_id,
        student.full_name,
        student.first_name,
        student.last_name,
        student.email,
    )



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#

import unittest
import random

import eyegrade.create.questions as questions


class _FixedQuestion(questions.FixedQuestion):
    def __init__(self, text):
        super().__init__(
            questions.QuestionVariation(
                questions.QuestionComponent(False, text=text), [], [], [], []
            )
        )


class TestQuestionsContainer(unittest.TestCase):
    def test_dunder_methods(self):
        container = questions.QuestionsContainer()
        container.append(_FixedQuestion("A"))
        container.append(
            questions.QuestionsGroup([_FixedQuestion("B"), _FixedQuestion("C")])
        )
        container.append(
            questions.QuestionsGroup(
                [_FixedQuestion("D"), _FixedQuestion("E"), _FixedQuestion("F")]
            )
        )
        container.append(_FixedQuestion("G"))
        self.assertEqual(len(container), 7)
        self.assertEqual(
            _extract_text(iter(container)), ["A", "B", "C", "D", "E", "F", "G"]
        )
        self.assertEqual(
            _extract_text(container[i] for i in range(7)),
            ["A", "B", "C", "D", "E", "F", "G"],
        )

    def test_shuffle(self):
        random.seed(1)
        # the first four calls to random.random() with this seed
        # should return 0.13, 0.85, 0.76, 0.26
        container = questions.QuestionsContainer()
        container.append(_FixedQuestion("A"))
        container.append(
            questions.QuestionsGroup([_FixedQuestion("B"), _FixedQuestion("C")])
        )
        container.append(
            questions.QuestionsGroup(
                [_FixedQuestion("D"), _FixedQuestion("E"), _FixedQuestion("F")]
            )
        )
        container.append(_FixedQuestion("G"))
        _, question_list, permutations = container.shuffle()
        self.assertEqual(
            _extract_text(question_list), ["A", "G", "D", "E", "F", "B", "C"]
        )
        self.assertEqual(permutations, [0, 6, 3, 4, 5, 1, 2])

    def test_positions(self):
        container = questions.QuestionsContainer()
        container.append(_FixedQuestion("A"))
        container.append(
            questions.QuestionsGroup([_FixedQuestion("B"), _FixedQuestion("C")])
        )
        container.append(
            questions.QuestionsGroup(
                [_FixedQuestion("D"), _FixedQuestion("E"), _FixedQuestion("F")]
            )
        )
        container.append(_FixedQuestion("G"))
        # pylint: disable=protected-access
        self.assertEqual(container._positions(container.groups[0]), [0])
        self.assertEqual(container._positions(container.groups[1]), [1, 2])
        self.assertEqual(container._positions(container.groups[2]), [3, 4, 5])
        self.assertEqual(container._positions(container.groups[3]), [6])


def _extract_text(question_list):
    return [question.text("").text for question in question_list]



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import cv2

from . import geometry
from . import utils
from . import scoring
from . import images

_color_blue = (255, 0, 0)
_color_good = (0, 210, 0)
_color_bad = (0, 0, 255)
_color_dot_bad = (255, 0, 0)
_color_dot_blank = (192, 0, 192)


class CellGeometry:
    def __init__(self, plu, pru, pld, prd, center, diagonal):
        """Receives the four corners of the cell, center point and diagonal.

        If center or diagonal are None, they are automatically computed
        from the values of the corners.

        """
        self.plu = plu
        self.pru = pru
        self.pld = pld
        self.prd = prd
        if center is not None:
            self.center = center
        else:
            self.center = geometry.rect_center(plu, pru, pld, prd)
        if diagonal is not None:
            self.diagonal = diagonal
        else:
            self.diagonal = geometry.distance(plu, prd)

    def corners(self):
        """Returns a tuple (plu, pru, pld, prd) with the cell corners."""
        return (self.plu, self.pru, self.pld, self.prd)


class ExamDecisions:
    def __init__(
        self, success, answers, detected_id, id_scores, model=None, infobits=None
    ):
        self.success = success
        self.answers = answers
        self.detected_id = detected_id
        self.id_scores = id_scores
        if model is not None:
            self.model = model
        elif infobits:
            self.model = utils.decode_model(infobits)
        else:
            self.model = None
        self.student = None
        self.students_rank = []

    def change_answer(self, question, answer):
        self.answers[question] = answer

    def set_students_rank(self, students_rank):
        self.students_rank = students_rank

    def set_student(self, student):
        self.student = student


class ExamCapture:
    """Capture of an exam, including the image, cell geometry and drawing."""

    def __init__(self, image, answer_cells, id_cells, progress=1.0):
        """Creates a new ExamCapture object.

        `image`: original capture of the exam (as captured by opencv);
        `answer_cells`: bi-dimensional list of num_questions x num_choices
                        CellGeometry objects. Each one represents an
                        answer cell.
        `id_cells`: list of num_digits CellGeometry objects. Each one
                    represents a digit cell for the student id (from
                    left to right).
        `progress`: progress ratio of the capture. Set 1.0 for exams in which
                    all the features have been detected.

        """
        self.image_raw = image
        self.image_drawn = None
        self.answer_cells = answer_cells
        self.id_cells = id_cells
        self.progress = progress
        self.reset_image()

    def has_answer_cells(self):
        return self.answer_cells is not None and len(self.answer_cells) > 0

    def has_id_cells(self):
        return self.id_cells is not None and len(self.id_cells) > 0

    def get_cell_clicked(self, point):
        """Determines the cell to which the given point corresponds.

        Returns (num_question, num_choice) or None if no cell corresponds.

        """
        min_dst = float("inf")
        num_question = None
        num_choice = None
        closest_cell = None
        for i, cells in enumerate(self.answer_cells):
            for j, cell in enumerate(cells):
                dst = geometry.distance(point, cell.center)
                if dst < min_dst:
                    min_dst = dst
                    num_question = i
                    num_choice = j + 1
                    closest_cell = cell
        if closest_cell is not None and min_dst <= closest_cell.diagonal / 2:
            return (num_question, num_choice)
        else:
            return (None, None)

    def reset_image(self):
        """Resets the drawn image by cloning the original image.

        All the drawings are lost.

        """
        if self.image_raw is not None:
            self.image_drawn = self.image_raw.copy()

    def save_image_drawn(self, filename):
        assert self.image_drawn is not None
        save_image(filename, self.image_drawn)

    def save_image_raw(self, filename):
        save_image(filename, self.image_raw)

    def draw_status(self):
        assert self.image_drawn is not None
        self._draw_status_bar()

    def draw_corner(self, point):
        assert self.image_drawn is not None
        cv2.circle(self.image_drawn, point, 4, _color_blue, thickness=1)

    def draw_answers(self, score):
        assert self.image_drawn is not None
        if score.answers:
            if score.solutions:
                self._draw_answers_solutions(score)
            else:
                self._draw_answers_no_solutions(score)

    def _draw_status_bar(self):
        x0 = images.get_width(self.image_drawn) - 60
        y0 = 10
        width = 50
        height = 20
        p0 = (x0, y0)
        p1 = geometry.round_point((x0 + self.progress * width, y0 + height))
        p2 = (x0 + width, y0 + height)
        cv2.rectangle(self.image_drawn, p0, p2, _color_blue)
        cv2.rectangle(self.image_drawn, p0, p1, _color_blue, thickness=-1)

    def _draw_answers_solutions(self, score):
        for answer, question_solutions, status, cells in zip(
            score.answers, score.solutions, score.answer_status, self.answer_cells
        ):
            if status == scoring.AnswerStatus.CORRECT:
                self._draw_cell_circle(cells[answer - 1], _color_good)
            elif status == scoring.AnswerStatus.INCORRECT:
                self._draw_cell_circle(cells[answer - 1], _color_bad)
                for solution in question_solutions:
                    self._draw_cell_center(cells[solution - 1], _color_dot_bad)
            elif status == scoring.AnswerStatus.BLANK:
                for solution in question_solutions:
                    self._draw_cell_center(cells[solution - 1], _color_dot_blank)
            elif status == scoring.AnswerStatus.VOID:
                self._draw_void_question(cells)

    def _draw_answers_no_solutions(self, score):
        for answer, cells in zip(score.answers, self.answer_cells):
            if answer > 0:
                self._draw_cell_circle(cells[answer - 1], _color_blue)

    def _draw_cell_circle(self, cell, color):
        radius = int(round(cell.diagonal / 3.5))
        cv2.circle(self.image_drawn, cell.center, radius, color, thickness=2)

    def _draw_cell_center(self, cell, color):
        cv2.circle(self.image_drawn, cell.center, 4, color, thickness=-1)

    def _draw_void_question(self, cells):
        cv2.line(
            self.image_drawn, cells[0].center, cells[-1].center, _color_bad, thickness=3
        )


def save_image(filename, image):
    cv2.imwrite(filename, image)



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import csv
import enum

import openpyxl

from . import utils


class FileFormat(enum.Enum):
    CSV_TABS = 1
    XLSX = 2


def create_writer(file_name, file_format):
    if file_format == FileFormat.CSV_TABS:
        writer = CSVWriter(file_name, utils.csv_tabs_dialect)
    elif file_format == FileFormat.XLSX:
        writer = XLSXWriter(file_name)
    else:
        raise ValueError("Unknown file format: {}".format(file_format))
    return writer


class XLSXWriter:
    def __init__(self, file_name):
        self.file_name = file_name

    def __enter__(self):
        self.workbook = openpyxl.Workbook()
        self.current_sheet = self.workbook.active
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        if traceback is None:
            self.workbook.save(self.file_name)
        self.current_sheet = None
        self.workbook = None

    def append_row(self, data):
        self.current_sheet.append(data)

    def append_sheet(self):
        self.current_sheet = self.workbook.create_sheet()

    def set_sheet_title(self, title):
        self.current_sheet.title = title

    def __str__(self):
        return "XLSXWriter({})".format(self.file_name)


class CSVWriter:
    def __init__(self, file_name, csv_dialect):
        self.file_name = file_name
        self.csv_dialect = csv_dialect

    def __enter__(self):
        self.file = open(self.file_name, "w")
        self.writer = csv.writer(self.file, dialect=self.csv_dialect)
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.file.close()

    def append_row(self, data):
        self.writer.writerow(data)

    def append_sheet(self):
        pass

    def set_sheet_title(self, title):
        pass

    def __str__(self):
        return "CSVWriter({})".format(self.file_name)


class GradesColumn:
    """Extract a field from a Student or Exam object, for exporting data"""

    STUDENT_KEYS = {
        "student_id": "Id",
        "name": "Name",
        "last_name": "Last name",
        "first_name": "First name",
    }
    EXAM_KEYS = {
        "exam_id": "Sequence number",
        "model": "Model",
        "correct": "Correct",
        "incorrect": "Incorrect",
        "score": "Score",
        "answers": None,
    }

    def __init__(self, column_key, num_questions=None):
        if column_key in GradesColumn.STUDENT_KEYS:
            self.extract = self._extract_from_student
            self.key = column_key
            self.headers = (GradesColumn.STUDENT_KEYS[column_key],)
        elif column_key in GradesColumn.EXAM_KEYS:
            self.extract = self._extract_from_exam
            self.key = column_key
            if column_key == "answers":
                if num_questions is not None:
                    self.headers = tuple(
                        "Q{}".format(i) for i in range(1, num_questions + 1)
                    )
                else:
                    raise ValueError("num_questions needs to be set")
            else:
                self.headers = (GradesColumn.EXAM_KEYS[column_key],)
        else:
            raise ValueError("Unknown column key: {}".format(column_key))

    def __str__(self):
        return "GradesColumn({})".format(self.key)

    def _extract_from_student(self, exam):
        return (getattr(exam["student"], self.key),)

    def _extract_from_exam(self, exam):
        if self.key == "answers":
            return exam["answers"]
        else:
            return (exam[self.key],)


class SortBy(enum.Enum):
    STUDENT_LIST = 1
    LAST_NAME = 2
    GRADING_SEQUENCE = 3


class GradesExportHelper:
    """Manage the options for exporting grades."""

    def __init__(self, exam_config, student_groups):
        self.student_groups = student_groups
        self.num_questions = exam_config.num_questions
        self.survey_mode = exam_config.survey_mode
        self.columns = []
        self.file_name = None
        self.file_format = None
        self.group = None
        self.one_sheet = None
        self.all_students = None
        self.sort_by = None
        self.add_column_headers = None
        self.export_by_exam = False

    def export_columns(self, keys):
        self.columns = [self._create_column(key) for key in keys]

    def export_group(self, index):
        self.group = self.student_groups[index]
        self.one_sheet = True
        self.export_by_exam = False

    def export_all_groups(self, one_sheet):
        self.group = None
        self.one_sheet = one_sheet
        self.export_by_exam = False

    def export_all_exams(self):
        self.group = None
        self.one_sheet = True
        self.export_by_exam = True

    def data(self, exam):
        data = []
        for column in self.columns:
            data.extend(column.extract(exam))
        return data

    def column_headers(self):
        data = []
        for column in self.columns:
            data.extend(column.headers)
        return data

    def iter_groups(self):
        if self.export_by_exam:
            yield (None, "Exams")
        elif self.group is None:
            if self.one_sheet:
                yield (None, "Group")
            else:
                for group in self.student_groups:
                    yield (group, group.name)
        else:
            yield (self.group, self.group.name)

    def create_writer(self):
        writer = create_writer(self.file_name, self.file_format)
        return writer

    def _create_column(self, key):
        return GradesColumn(key, num_questions=self.num_questions)



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import numpy as np

from . import sample


class Evaluation:
    def __init__(self, classifier, samples):
        self.classifier = classifier
        self.samples = samples
        self._evaluate()

    @property
    def confusion_matrix_r(self):
        return (
            np.array(self.confusion_matrix, dtype="float32")
            / np.sum(self.confusion_matrix, axis=1)[np.newaxis].T
        )

    @property
    def success_rate_balanced(self):
        return np.mean(self.confusion_matrix_r.diagonal())

    def _evaluate(self):
        num_classes = self.classifier.num_classes
        self.results = np.zeros(len(self.samples), dtype=bool)
        self.confusion_matrix = np.zeros(shape=(num_classes, num_classes), dtype="int")
        for i, samp in enumerate(self.samples):
            detected = self.classifier.classify(samp)
            self.confusion_matrix[samp.label, detected] += 1
            self.results[i] = samp.check_label(detected)
        self.success_rate = sum(self.results) / len(self.results)


class KFoldCrossEvaluation(Evaluation):
    def __init__(
        self,
        classifier,
        sample_sets,
        oversampling=False,
        training_params=None,
        threshold=None,
    ):
        self.classifier = classifier
        self.sample_sets = sample_sets
        self.training_params = training_params
        self.threshold = threshold
        self._evaluate(oversampling=oversampling)

    def _evaluate(self, oversampling=False):
        num_classes = self.classifier.num_classes
        self.confusion_matrix = np.zeros(shape=(num_classes, num_classes), dtype="int")
        for i, evaluation_set in enumerate(self.sample_sets):
            training_set = sample.SampleSet()
            training_set.load_from_sample_sets(self.sample_sets[:i])
            training_set.load_from_sample_sets(self.sample_sets[i + 1 :])
            if oversampling:
                training_set = training_set.oversample()
            self.classifier.train(training_set.samples(), params=self.training_params)
            evaluation = Evaluation(self.classifier, evaluation_set)
            self.confusion_matrix += evaluation.confusion_matrix
            self.classifier.reset()
            total = self.confusion_matrix.sum()
            correct = self.confusion_matrix.diagonal().sum()
            self.success_rate = correct / total
            print("Round {}: {}".format(i, self.success_rate))
            if self.threshold is not None and self.success_rate < self.threshold:
                break



import unittest
import os.path


def _load_tests(loader, standard_tests, pattern):
    this_dir = os.path.dirname(__file__)
    package_tests = loader.discover(start_dir=this_dir, pattern=pattern)
    standard_tests.addTests(package_tests)
    return standard_tests


def get_tests():
    return _load_tests(unittest.TestLoader(), unittest.TestSuite(), "test*.py")



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
import os
import unittest

import numpy as np

import eyegrade.ocr.sample as sample
import eyegrade.ocr.classifiers as classifiers


class TestClassifier(unittest.TestCase):
    def _get_test_file_path(self, filename):
        dirname = os.path.dirname(os.path.abspath(__file__))
        return os.path.join(dirname, filename)

    def test_classify_digit(self):
        image_path = self._get_test_file_path("digit.png")
        corners = np.array([[0, 1], [19, 0], [0, 7], [21, 17]])
        samp = sample.Sample(corners, image_filename=image_path)
        classifier = classifiers.DefaultDigitClassifier()
        label = classifier.classify(samp)
        self.assertTrue(label in range(10))

    def test_classify_cross(self):
        image_path = self._get_test_file_path("cross.png")
        corners = np.array([[0, 0], [27, 0], [1, 32], [29, 32]])
        samp = sample.Sample(corners, image_filename=image_path)
        classifier = classifiers.DefaultCrossesClassifier()
        label = classifier.classify(samp)
        self.assertTrue(label == 0 or label == 1)



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
import unittest
import fractions
import decimal

import eyegrade.scoring as scoring
import eyegrade.exams as exams


class TestQuestionScores(unittest.TestCase):
    def testFloat(self):
        score = scoring.QuestionScores(1.0, 0.5, 0.0)
        key_1 = "1.0000000000000000"
        key_2 = "0.5000000000000000"
        key_3 = "0.0000000000000000"
        self.assertEqual(score.format_all(), ";".join((key_1, key_2, key_3)))
        self.assertEqual(score.format_score(scoring.AnswerStatus.CORRECT), key_1)
        self.assertEqual(score.format_correct_score(), key_1)
        self.assertEqual(score.format_score(scoring.AnswerStatus.INCORRECT), key_2)
        self.assertEqual(score.format_incorrect_score(), key_2)
        self.assertEqual(score.format_score(scoring.AnswerStatus.BLANK), key_3)
        self.assertEqual(score.format_blank_score(), key_3)
        self.assertEqual(score.score(scoring.AnswerStatus.CORRECT), 1.0)
        self.assertEqual(score.score(scoring.AnswerStatus.INCORRECT), -0.5)
        self.assertEqual(score.score(scoring.AnswerStatus.BLANK), 0.0)
        self.assertEqual(score.format_weight(), "1")

    def testDecimal(self):
        score = scoring.QuestionScores(
            decimal.Decimal("1.0"), decimal.Decimal("0.5"), decimal.Decimal("0.0")
        )
        key_1 = "1.0"
        key_2 = "0.5"
        key_3 = "0.0"
        self.assertEqual(score.format_all(), ";".join((key_1, key_2, key_3)))
        self.assertEqual(score.format_score(scoring.AnswerStatus.CORRECT), key_1)
        self.assertEqual(score.format_correct_score(), key_1)
        self.assertEqual(score.format_score(scoring.AnswerStatus.INCORRECT), key_2)
        self.assertEqual(score.format_incorrect_score(), key_2)
        self.assertEqual(score.format_score(scoring.AnswerStatus.BLANK), key_3)
        self.assertEqual(score.format_blank_score(), key_3)
        self.assertEqual(score.score(scoring.AnswerStatus.CORRECT), 1.0)
        self.assertEqual(score.score(scoring.AnswerStatus.INCORRECT), -0.5)
        self.assertEqual(score.score(scoring.AnswerStatus.BLANK), 0.0)
        self.assertEqual(score.format_weight(), "1")

    def testDecimal2(self):
        score = scoring.QuestionScores("1.0", "0.5", "0.0")
        key_1 = "1.0"
        key_2 = "0.5"
        key_3 = "0.0"
        self.assertEqual(score.format_all(), ";".join((key_1, key_2, key_3)))
        self.assertEqual(score.format_score(scoring.AnswerStatus.CORRECT), key_1)
        self.assertEqual(score.format_correct_score(), key_1)
        self.assertEqual(score.format_score(scoring.AnswerStatus.INCORRECT), key_2)
        self.assertEqual(score.format_incorrect_score(), key_2)
        self.assertEqual(score.format_score(scoring.AnswerStatus.BLANK), key_3)
        self.assertEqual(score.format_blank_score(), key_3)
        self.assertEqual(score.score(scoring.AnswerStatus.CORRECT), 1.0)
        self.assertEqual(score.score(scoring.AnswerStatus.INCORRECT), -0.5)
        self.assertEqual(score.score(scoring.AnswerStatus.BLANK), 0.0)
        self.assertEqual(score.format_weight(), "1")

    def testFractionAndInt(self):
        score = scoring.QuestionScores("1", "1/3", "0")
        self.assertEqual(score.format_all(), "1;1/3;0")
        self.assertEqual(score.format_score(scoring.AnswerStatus.CORRECT), "1")
        self.assertEqual(score.format_score(scoring.AnswerStatus.INCORRECT), "1/3")
        self.assertEqual(score.format_score(scoring.AnswerStatus.BLANK), "0")
        self.assertEqual(score.score(scoring.AnswerStatus.CORRECT), 1)
        self.assertEqual(
            score.score(scoring.AnswerStatus.INCORRECT), fractions.Fraction(-1, 3)
        )
        self.assertEqual(score.score(scoring.AnswerStatus.BLANK), 0)
        self.assertEqual(score.format_weight(), "1")

    def testSignedFormat(self):
        score = scoring.QuestionScores("1", "1/3", "1/6")
        result = score.format_score(scoring.AnswerStatus.CORRECT, signed=True)
        self.assertEqual(result, "1")
        result = score.format_score(scoring.AnswerStatus.INCORRECT, signed=True)
        self.assertEqual(result, "-1/3")
        result = score.format_score(scoring.AnswerStatus.BLANK, signed=True)
        self.assertEqual(result, "-1/6")
        result = score.format_correct_score(signed=True)
        self.assertEqual(result, "1")
        result = score.format_incorrect_score(signed=True)
        self.assertEqual(result, "-1/3")
        result = score.format_blank_score(signed=True)
        self.assertEqual(result, "-1/6")
        score = scoring.QuestionScores("1.0", "0.5", "0.25")
        result = score.format_score(scoring.AnswerStatus.CORRECT, signed=True)
        self.assertEqual(result, "1.0")
        result = score.format_score(scoring.AnswerStatus.INCORRECT, signed=True)
        self.assertEqual(result, "-0.5")
        result = score.format_score(scoring.AnswerStatus.BLANK, signed=True)
        self.assertEqual(result, "-0.25")

    def testWeight(self):
        score = scoring.QuestionScores("1", "1/3", "1/6", weight="3/2")
        self.assertEqual(score.format_all(), "1;1/3;1/6")
        self.assertEqual(score.format_score(scoring.AnswerStatus.CORRECT), "1")
        self.assertEqual(score.format_score(scoring.AnswerStatus.INCORRECT), "1/3")
        self.assertEqual(score.format_score(scoring.AnswerStatus.BLANK), "1/6")
        self.assertEqual(
            score.score(scoring.AnswerStatus.CORRECT), fractions.Fraction(3, 2)
        )
        self.assertEqual(
            score.score(scoring.AnswerStatus.INCORRECT), fractions.Fraction(-1, 2)
        )
        self.assertEqual(
            score.score(scoring.AnswerStatus.BLANK), fractions.Fraction(-1, 4)
        )
        self.assertEqual(score.format_weight(), "3/2")

    def testNegativeScores(self):
        self.assertRaises(ValueError, scoring.QuestionScores, "-1/3", "1/6", "0")
        score = scoring.QuestionScores("1", "-1/3", "1/6")
        self.assertEqual(
            score.score(scoring.AnswerStatus.CORRECT), fractions.Fraction(1, 1)
        )
        self.assertEqual(
            score.score(scoring.AnswerStatus.INCORRECT), fractions.Fraction(-1, 3)
        )
        self.assertEqual(
            score.score(scoring.AnswerStatus.BLANK), fractions.Fraction(-1, 6)
        )
        score = scoring.QuestionScores("1", "1/3", "-1/6")
        self.assertEqual(
            score.score(scoring.AnswerStatus.CORRECT), fractions.Fraction(1, 1)
        )
        self.assertEqual(
            score.score(scoring.AnswerStatus.INCORRECT), fractions.Fraction(-1, 3)
        )
        self.assertEqual(
            score.score(scoring.AnswerStatus.BLANK), fractions.Fraction(-1, 6)
        )

    def testNegativeWeights(self):
        self.assertRaises(
            ValueError, scoring.QuestionScores, "1/3", "1/6", "0", weight="-1/2"
        )
        self.assertRaises(
            ValueError, scoring.QuestionScores, "1/3", "1/6", "0", weight="-1"
        )
        self.assertRaises(
            ValueError, scoring.QuestionScores, "1/3", "1/6", "0", weight="-0.5"
        )

    def testBadValues(self):
        self.assertRaises(ValueError, scoring.QuestionScores, "1//3", "1/6", "0")
        self.assertRaises(ValueError, scoring.QuestionScores, "1", "1a", "0")
        self.assertRaises(ValueError, scoring.QuestionScores, "1", "2", "0.z3")

    def testClone(self):
        score1 = scoring.QuestionScores("1", "1/3", "1/6", weight="3/2")
        score2 = score1.clone()
        self.assertEqual(score2.correct_score, score1.correct_score)
        self.assertEqual(score2.incorrect_score, score1.incorrect_score)
        self.assertEqual(score2.blank_score, score1.blank_score)
        self.assertEqual(score2.weight, score1.weight)
        score3 = score1.clone(new_weight=2)
        self.assertEqual(score3.correct_score, score1.correct_score)
        self.assertEqual(score3.incorrect_score, score1.incorrect_score)
        self.assertEqual(score3.blank_score, score1.blank_score)
        self.assertEqual(score3.weight, 2)

    def testSort(self):
        scores = [
            scoring.QuestionScores("1", "1/3", "0"),
            scoring.QuestionScores("2", "2/3", "1/2"),
            scoring.QuestionScores("1", "1/3", "0"),
            scoring.QuestionScores("2", "2/3", "0"),
            scoring.QuestionScores("1/2", "1/6", "0"),
        ]
        sorted_scores = [scores[4], scores[0], scores[2], scores[3], scores[1]]
        self.assertEqual(sorted(scores), sorted_scores)


class TestExamConfigScores(unittest.TestCase):
    def testSetQuestionScores(self):
        exam = exams.ExamConfig()
        exam.num_questions = 5
        scores = [
            scoring.QuestionScores("1", "1/3", "0"),
            scoring.QuestionScores("2", "2/3", "0"),
            scoring.QuestionScores("1", "1/3", "0"),
            scoring.QuestionScores("2", "2/3", "0"),
            scoring.QuestionScores("1/2", "1/6", "0"),
        ]
        exam.set_question_scores("A", scores)
        exam.set_question_scores(
            "B", [scores[1], scores[2], scores[4], scores[0], scores[3]]
        )

    def testSetQuestionScoresError(self):
        exam = exams.ExamConfig()
        scores = [
            scoring.QuestionScores("1", "1/3", "0"),
            scoring.QuestionScores("2", "2/3", "0"),
            scoring.QuestionScores("1", "1/3", "0"),
            scoring.QuestionScores("2", "2/3", "0"),
            scoring.QuestionScores("1/2", "1/6", "0"),
        ]
        self.assertRaises(ValueError, exam.set_question_scores, "A", scores)
        exam.num_questions = 5
        exam.set_question_scores("A", scores)
        self.assertRaises(
            ValueError,
            exam.set_question_scores,
            "B",
            [scores[1], scores[1], scores[4], scores[0], scores[3]],
        )

    def testSetEqualScores1(self):
        exam = exams.ExamConfig()
        exam.num_questions = 5
        exam.set_base_scores(scoring.QuestionScores("1", "1/2", "0"))
        exam.set_equal_scores("A")
        exam.set_equal_scores("B")
        for scores in exam.scores["A"] + exam.scores["B"]:
            self.assertEqual(scores, exam.base_scores)

    def testSetEqualScores2(self):
        exam = exams.ExamConfig()
        exam.num_questions = 5
        exam.models = ["A", "B"]
        exam.set_base_scores(scoring.QuestionScores("1", "1/2", "0"), same_weights=True)
        for scores in exam.scores["A"] + exam.scores["B"]:
            self.assertEqual(scores, exam.base_scores)

    def testSetEqualScores3(self):
        """Test for issue #96.

        An exception because of different weights was raised because
        of the issue.

        """
        exam = exams.ExamConfig()
        exam.num_questions = 5
        exam.set_base_scores(scoring.QuestionScores("1", "1/2", "0"), same_weights=True)
        exam.set_equal_scores("A")
        exam.set_equal_scores("B")
        for scores in exam.scores["A"] + exam.scores["B"]:
            self.assertEqual(scores, exam.base_scores)
        exam.set_base_scores(scoring.QuestionScores("2", "1", "0"), same_weights=True)
        for scores in exam.scores["A"] + exam.scores["B"]:
            self.assertEqual(scores, exam.base_scores)

    def testSetBaseScoresError(self):
        exam = exams.ExamConfig()
        scores = scoring.QuestionScores("1", "1/2", "0", weight="2")
        self.assertRaises(ValueError, exam.set_base_scores, scores)

    def testSetQuestionWeights(self):
        exam = exams.ExamConfig()
        exam.num_questions = 5
        scores = [
            scoring.QuestionScores("1", "1/2", "0", weight="1"),
            scoring.QuestionScores("1", "1/2", "0", weight="2"),
            scoring.QuestionScores("1", "1/2", "0", weight="1"),
            scoring.QuestionScores("1", "1/2", "0", weight="2"),
            scoring.QuestionScores("1", "1/2", "0", weight="1/2"),
            scoring.QuestionScores("1", "1/2", "0", weight="1"),
            scoring.QuestionScores("1", "1/2", "0", weight="1"),
            scoring.QuestionScores("1", "1/2", "0", weight="1/2"),
            scoring.QuestionScores("1", "1/2", "0", weight="2"),
            scoring.QuestionScores("1", "1/2", "0", weight="2"),
        ]
        exam.set_base_scores(scoring.QuestionScores("1", "1/2", "0"))
        exam.set_question_weights("A", [1, 2, 1, 2, "1/2"])
        exam.set_question_weights("B", ["1", "1", "1/2", "2", "2"])
        for value, key in zip(exam.scores["A"] + exam.scores["B"], scores):
            self.assertEqual(value, key)

    def testSetQuestionWeightsError(self):
        exam = exams.ExamConfig()
        self.assertRaises(ValueError, exam.set_question_weights, "A", [1, 2, 2])
        exam.num_questions = 3
        self.assertRaises(ValueError, exam.set_question_weights, "A", [1, 2, 2])
        exam.set_base_scores(scoring.QuestionScores("1", "1/3", "0"))
        exam.set_question_weights("A", [1, 2, 2])
        self.assertRaises(ValueError, exam.set_question_weights, "B", [2, 1, 1])

    def testGetQuestionWeights(self):
        exam = exams.ExamConfig()
        exam.num_questions = 3
        self.assertEqual(exam.get_question_weights("A"), None)
        exam.set_base_scores(scoring.QuestionScores("1", "1/2", "0"))
        self.assertEqual(exam.get_question_weights("A"), None)
        exam.set_question_weights("A", [1, 2, "1/2"])
        exam.set_question_weights("B", ["1/2", "1", "2"])
        self.assertEqual(
            exam.get_question_weights("A"), [1, 2, fractions.Fraction(1, 2)]
        )
        self.assertEqual(
            exam.get_question_weights("B"), [fractions.Fraction(1, 2), 1, 2]
        )
        self.assertEqual(
            exam.get_question_weights("A", formatted=True), ["1", "2", "1/2"]
        )
        self.assertEqual(
            exam.get_question_weights("B", formatted=True), ["1/2", "1", "2"]
        )

    def testAllWeightsAreOneWeights(self):
        exam = exams.ExamConfig()
        exam.num_questions = 5
        self.assertFalse(exam.all_weights_are_one())
        exam.models = ["A", "B"]
        self.assertFalse(exam.all_weights_are_one())
        exam.set_base_scores(scoring.QuestionScores("1", "1/2", "0"), same_weights=True)
        self.assertTrue(exam.all_weights_are_one())

    def testAllWeightsAreOneWeightsNegative(self):
        exam = exams.ExamConfig()
        exam.num_questions = 5
        exam.set_base_scores(scoring.QuestionScores("1", "1/2", "0"))
        exam.set_question_weights("A", [1, 2, 1, "1/2", 1])
        exam.set_question_weights("B", [2, 1, 1, 1, "1/2"])
        self.assertFalse(exam.all_weights_are_one())


class TestScore(unittest.TestCase):
    def testScoreNoScores(self):
        answers = [0, 1, 2, 3, 0, 1]
        solutions = [{1}, {1}, {3}, {3}, {4}, {1}]
        question_scores = None
        score = scoring.Score(answers, solutions, question_scores)
        self.assertEqual(score.correct, 3)
        self.assertEqual(score.incorrect, 1)
        self.assertEqual(score.blank, 2)
        self.assertEqual(score.score, None)
        self.assertEqual(score.max_score, None)

    def testScoreEqualScores(self):
        answers = [0, 1, 2, 3, 0, 1]
        solutions = [{1}, {1}, {3}, {3}, {4}, {1}]
        base_score = scoring.QuestionScores("1", "1/2", "0")
        question_scores = 6 * [base_score]
        score = scoring.Score(answers, solutions, question_scores)
        self.assertEqual(score.correct, 3)
        self.assertEqual(score.incorrect, 1)
        self.assertEqual(score.blank, 2)
        self.assertEqual(score.score, 2.5)
        self.assertEqual(score.max_score, 6.0)

    def testScoreDifferentScores(self):
        answers = [0, 1, 2, 3, 0, 1]
        solutions = [{1}, {1}, {3}, {3}, {4}, {1}]
        base_score = scoring.QuestionScores("1", "1/2", "0")
        question_scores = [
            base_score,
            base_score.clone(new_weight=2),
            base_score.clone(new_weight=2),
            base_score,
            base_score,
            base_score.clone(new_weight=3),
        ]
        score = scoring.Score(answers, solutions, question_scores)
        self.assertEqual(score.correct, 3)
        self.assertEqual(score.incorrect, 1)
        self.assertEqual(score.blank, 2)
        self.assertEqual(score.score, 5.0)
        self.assertEqual(score.max_score, 10.0)

    def testScoreError(self):
        answers = [0, 1, 2, 3, 0, 1]
        solutions = [{1}, {1}, {3}, {3}, {4}, {1}]
        base_score = scoring.QuestionScores("1", "1/2", "0")
        question_scores = [
            base_score,
            base_score.clone(new_weight=2),
            base_score.clone(new_weight=2),
            base_score,
            base_score,
        ]
        self.assertRaises(
            ValueError, scoring.Score, answers, solutions, question_scores
        )

    def testScoreMultipleCorrect(self):
        answers = [1, 1, 2, 3, 2, 1]
        solutions = [{1}, {1}, {2, 3}, {3}, {4}, {1}]
        base_score = scoring.QuestionScores("1", "0", "0")
        question_scores = 6 * [base_score]
        score = scoring.Score(answers, solutions, question_scores)
        self.assertEqual(score.correct, 5)
        self.assertEqual(score.incorrect, 1)
        self.assertEqual(score.blank, 0)
        self.assertEqual(score.score, 5.0)
        self.assertEqual(score.max_score, 6.0)

    def testDecimalScores(self):
        answers = [1, 1]
        solutions = [{1}, {3}]
        question_scores = [
            scoring.QuestionScores("2.75", "1.25", "0"),
            scoring.QuestionScores("2.75", "1.25", "0"),
        ]
        self.assertTrue(isinstance(question_scores[0].correct_score, decimal.Decimal))
        score = scoring.Score(answers, solutions, question_scores)
        self.assertEqual(score.correct, 1)
        self.assertEqual(score.incorrect, 1)
        self.assertEqual(score.blank, 0)
        self.assertAlmostEqual(score.score, 1.5)
        self.assertAlmostEqual(score.max_score, 5.5)


class TestAutomaticScore(unittest.TestCase):
    def testAutomaticScoreDecimals(self):
        automatic_score = scoring.AutomaticScore("2.5", True)
        question_scores = automatic_score.compute(10, 3)
        self.assertEqual(question_scores.correct_score, decimal.Decimal("0.25"))
        self.assertEqual(question_scores.incorrect_score, decimal.Decimal("0.125"))
        self.assertEqual(question_scores.blank_score, 0)

    def testAutomaticScoreFractions(self):
        automatic_score = scoring.AutomaticScore("5/2", True)
        question_scores = automatic_score.compute(10, 3)
        self.assertEqual(question_scores.correct_score, fractions.Fraction(1, 4))
        self.assertEqual(question_scores.incorrect_score, fractions.Fraction(1, 8))
        self.assertEqual(question_scores.blank_score, 0)


if __name__ == "__main__":
    unittest.main()



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import gettext

from PyQt6.QtGui import QIcon, QImage, QPainter, QPixmap, QRegularExpressionValidator

from PyQt6.QtWidgets import (
    QButtonGroup,
    QCheckBox,
    QComboBox,
    QCompleter,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QRadioButton,
    QSpinBox,
    QStatusBar,
    QStyle,
    QTableView,
    QWidget,
)

from PyQt6.QtCore import (
    QAbstractListModel,
    QModelIndex,
    QRegularExpression,
    QSortFilterProxyModel,
    Qt,
)

# Typing: it seems that pyqtProperty isn't supported by PyQt6 stubs (Feb 2022)
from PyQt6.QtCore import pyqtProperty  # type: ignore

from .. import utils
from .. import scoring
from . import Colors

t = gettext.translation("eyegrade", utils.locale_dir(), fallback=True)
_ = t.gettext


class LineContainer(QWidget):
    """Container that disposes other widgets horizontally."""

    def __init__(self, parent, *widgets):
        super().__init__(parent)
        self.layout = QHBoxLayout(self)
        self.setLayout(self.layout)
        for widget in widgets:
            self.add(widget)

    def add(self, widget):
        self.layout.addWidget(widget)


class CompletingComboBox(QComboBox):
    """An editable combo box that filters and autocompletes."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setEditable(True)
        self.filter = QSortFilterProxyModel(self)
        self.filter.setFilterCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self.filter.setSourceModel(self.model())
        self.completer = QCompleter(self.filter, self)
        self.completer.setCompletionMode(
            QCompleter.CompletionMode.UnfilteredPopupCompletion
        )
        self.setCompleter(self.completer)
        self.lineEdit().textEdited.connect(self.filter.setFilterFixedString)
        self.currentIndexChanged.connect(self._index_changed)

    def _index_changed(self, index):
        self.lineEdit().selectAll()


class StudentComboBox(CompletingComboBox):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.lineEdit().selectAll()
        self.lineEdit().setFocus()
        self.students = []

    def add_students(self, students):
        for student in students:
            self.add_student(student)

    def add_student(self, student, set_current=False):
        self.addItem(student.id_and_name)
        self.students.append(student)
        if set_current:
            self.setCurrentIndex(len(self.students) - 1)

    def current_student(self):
        student = None
        index = self.currentIndex()
        if index >= 0:
            if index < len(self.students):
                if self.currentText() == self.students[index].id_and_name:
                    # The user hasn't edited the text of this item
                    student = self.students[index]
            if student is None and self.currentText():
                # Perhaps a new entry is selected but the text is the
                # same as in an existing entry
                try:
                    index = [s.id_and_name for s in self.students].index(
                        self.currentText()
                    )
                except ValueError:
                    # Not in the student list
                    pass
                else:
                    self.setCurrentIndex(index)
                    student = self.students[index]
        return student


class StatusBar(QStatusBar):
    """Status bar for the main window.

    For now it just contains a simple QLabel.

    """

    def __init__(self, parent):
        """Creates a new instance.

        :param parent: The parent of this status bar.

        """
        super().__init__(parent=parent)
        self.status_label = QLabel(parent=self)
        self.addWidget(self.status_label)
        self._show_program_version()
        self.setStyleSheet("QStatusBar {border-top: 1px solid " "rgb(128, 128, 128); }")

    def set_message(self, text):
        """Sets a new left-side status text.

        :param str text: The text to display in the status bar.

        """
        self.status_label.setText(text)

    def _show_program_version(self):
        version_line = '{0} {1} - <a href="{2}">{2}</a>'.format(
            utils.program_name, utils.version, utils.web_location
        )
        label = QLabel(version_line)
        label.setOpenExternalLinks(True)
        self.addPermanentWidget(label)


class LabelledCheckBox(QWidget):
    """A checkbox with a label."""

    def __init__(self, label_text, parent, checked=False, enabled=True):
        """Creates a new instance.

        :param label: The label to show with the checkbox.
        :param parent: The parent of this widget.
        :param checked: Initial state of the checkbox (defaults to False).

        """
        super().__init__(parent=parent)
        layout = QHBoxLayout(self)
        self.checkbox = QCheckBox(parent=self)
        self.checkbox.setChecked(checked)
        self.checkbox.setEnabled(enabled)
        layout.addWidget(self.checkbox, alignment=Qt.AlignmentFlag.AlignLeft)
        layout.addWidget(
            QLabel(label_text, parent=self),
            stretch=1,
            alignment=Qt.AlignmentFlag.AlignLeft,
        )
        self.setLayout(layout)

    def is_checked(self):
        """Returns True if the checkbox is checked, False otherwise."""
        return self.checkbox.isChecked()


class OpenFileWidget(QWidget):
    """Dialog with a text field and a button to open a file selector."""

    def __init__(
        self,
        parent,
        select_directory=False,
        name_filter="",
        minimum_width=200,
        title="",
        check_file_function=None,
    ):
        super().__init__(parent)
        self.select_directory = select_directory
        self.name_filter = name_filter
        self.title = title
        self._check_file = check_file_function
        layout = QHBoxLayout(self)
        self.setLayout(layout)
        self.filename_widget = QLineEdit(self)
        self.filename_widget.setMinimumWidth(minimum_width)
        self.button = QPushButton(
            QIcon(utils.resource_path("open_file.svg")), "", parent=self
        )
        self.button.clicked.connect(self._open_dialog)
        layout.addWidget(self.filename_widget)
        layout.addWidget(self.button)
        layout.setContentsMargins(0, 0, 0, 0)
        self.last_validated_value = None

    def text(self):
        return self.filename_widget.text()

    def set_text(self, filename):
        self.filename_widget.setText(filename)

    def setEnabled(self, enabled):
        """Toggle enabled status of this widget.

        If the widget is disabled, the validated status
        is forced to True with the statement
        self.last_validated_value = self.text()

        """
        self.filename_widget.setEnabled(enabled)
        self.button.setEnabled(enabled)
        if not enabled:
            self.last_validated_value = self.text()
        else:
            self.last_validated_value = None

    def is_validated(self):
        """Returns True if the value equals the latest validated value.

        This way, the file needs not to be validated again if it has not been
        changed since the last validation.

        """
        return self.last_validated_value == self.text()

    def check_value(self, filename=None):
        """Checks the file and returns True if it is valid.

        If it is not valid, shows an error message. It the validation function
        has not been set (it is None), returns always True.

        If `filename` is None, the internal value is used instead.

        """
        if filename is None:
            filename = self.text()
        valid = True
        if self._check_file is not None:
            valid, msg = self._check_file(filename)
        if not valid:
            QMessageBox.critical(self, _("Error"), msg)
        else:
            self.last_validated_value = filename
        return valid

    def _open_dialog(self, value):
        if self.select_directory:
            filename = QFileDialog.getExistingDirectory(
                self,
                self.title,
                "",
                (
                    QFileDialog.Option.ShowDirsOnly
                    | QFileDialog.Option.DontResolveSymlinks
                    | QFileDialog.Option.DontUseNativeDialog
                ),
            )
        else:
            filename, _ = QFileDialog.getOpenFileName(
                self,
                self.title,
                "",
                self.name_filter,
                None,
                QFileDialog.Option.DontUseNativeDialog,
            )
        if filename:
            valid = self.check_value(filename=filename)
            if valid:
                self.filename_widget.setText(filename)


class InputScore(QLineEdit):
    """Allows the user to enter a score."""

    def __init__(self, parent=None, minimum_width=100, is_positive=True):
        super().__init__(parent=parent)
        self.setMinimumWidth(minimum_width)
        self.is_positive = is_positive
        if is_positive:
            placeholder = _("e.g.: 2; 2.5; 5/2")
        else:
            placeholder = _("e.g.: 0; -1; -1.25; -5/4")
        self.setPlaceholderText(placeholder)
        regex = r"((\d*(\.\d+))|(\d+\/\d+))"
        if not is_positive:
            regex = "-?" + regex
        validator = QRegularExpressionValidator(QRegularExpression(regex), self)
        self.setValidator(validator)

    def value(self, force_float=False):
        """Returns the value as a fractions.Fraction, a decimal.Decimal or a float.

        Returns None if the field is empty or the value is not
        correct.  If `force_float` a float is returned always.

        """
        allow_negatives = not self.is_positive
        try:
            score = scoring.parse_number(
                self.text(), force_float=force_float, allow_negatives=allow_negatives
            )
        except ValueError:
            score = None
        return score

    def setPlaceholderText(self, text):
        """Proxy for the same method in QLineEdit.

        This method is overridden because some old versions of Qt4 do
        not provide the method. This proxy method just calls the one
        from QLineEdit, and fails silently if the method does not
        exist there.

        """
        try:
            super().setPlaceholderText(text)
        except AttributeError:
            # Just do nothing if the version of Qt/PyQt is old...
            pass


class CamView(QWidget):
    def __init__(self, size, parent, draw_logo=False, border=False):
        super().__init__(parent)
        if not border:
            fixed_size = size
        else:
            fixed_size = (size[0] + 10, size[1] + 10)
        self.setFixedSize(*fixed_size)
        self.border = border
        self.image_size = size
        self.display_wait_image()
        if draw_logo:
            self.logo = QPixmap(utils.resource_path("logo.svg"))
        else:
            self.logo = None
        self.mouse_listener = None

    def paintEvent(self, event):
        painter = QPainter(self)
        if self.border:
            size = self.size()
            painter.setPen(Colors.eyegrade_blue)
            painter.drawRoundedRect(0, 0, size.width() - 2, size.height() - 2, 10, 10)
            painter.drawImage(5, 5, self.image)
        else:
            painter.drawImage(event.rect(), self.image)

    def display_capture(self, cv_image):
        """Displays a captured image in the window.

        The image is in the numpy format used by opencv.

        """
        # It is important to use the variable data to prevent issue #58.
        data = cv_image.data
        height, width, nbytes = cv_image.shape
        self.image = QImage(
            data, width, height, nbytes * width, QImage.Format.Format_RGB888
        ).rgbSwapped()
        if self.logo is not None:
            painter = QPainter(self.image)
            painter.drawPixmap(width - 40, height - 40, 36, 36, self.logo)
        self.update()

    def display_wait_image(self):
        self.image = QImage(
            self.image_size[0], self.image_size[1], QImage.Format.Format_RGB888
        )
        self.image.fill(Qt.GlobalColor.darkBlue)
        self.update()

    def register_mouse_pressed_listener(self, listener):
        """Registers a function to receive a mouse clicked event.

        The listener must receive as parameter a tuple (x, y).

        """
        self.mouse_listener = listener

    def mousePressEvent(self, event):
        if self.mouse_listener:
            self.mouse_listener((int(event.position().x()), int(event.position().y())))


class InputCustomPattern(QLineEdit):
    """Allows the user to enter a string with a specific pattern validation.

    The pattern is a regular expression.

    """

    def __init__(self, parent=None, fixed_size=40, regex=r".+", placeholder=None):
        super().__init__(parent=parent)
        if placeholder:
            self.setPlaceholderText(placeholder)
        self.setFixedWidth(fixed_size)
        validator = QRegularExpressionValidator(QRegularExpression(regex), self)
        self.setValidator(validator)


class InputInteger(QSpinBox):
    """Allows the user to enter an integer field"""

    def __init__(self, parent=None, initial_value=1, min_value=1, max_value=100):
        super().__init__(parent=parent)
        self.setRange(min_value, max_value)
        self.setValue(initial_value)


class InputRadioGroup(QWidget):
    """Create an horizontal radio group"""

    def __init__(self, parent=None, option_list=None, default_select=0):
        super().__init__(parent=parent)
        layout = QHBoxLayout(self)
        self.group = QButtonGroup()
        for idx, op in enumerate(option_list):
            self.op = QRadioButton(_(op))
            if idx == default_select:
                self.op.setChecked(True)
            layout.addWidget(self.op)
            self.group.addButton(self.op)
        self.setLayout(layout)

    @pyqtProperty(str)
    def currentItemData(self):
        return str(abs(int(self.group.checkedId())) - 1)


class ItemList:
    """Custom item for permutation list"""

    def __init__(self, optionName, optionNumber):
        super().__init__()
        self.name = optionName
        self.numb = optionNumber
        self.perm = {}

    def get_question_number(self):
        return str(self.numb)

    def get_permutation(self):
        return self.perm

    def set_permutation(self, permutation):
        self.perm = permutation
        return True


class InputComboBox(QComboBox):
    """A Combobox with a specific ID"""

    def __init__(self, parent=None, c_type=None, form=0, alternative=0):
        super().__init__(parent=parent)
        self.c_type = c_type
        self.form = form
        self.alternative = alternative


class CustomTableView(QTableView):
    """QTableView that can compute its own size."""

    def __init__(self, parent=None, maximum_width=500, maximum_height=300):
        super().__init__(parent=parent)
        self.maximum_height = maximum_height
        self.maximum_width = maximum_width

    def adjust_size(self):
        self.resizeRowsToContents()
        self.resizeColumnsToContents()
        self._set_dimension_hints()

    def adjust_columns_size(self):
        self.resizeColumnsToContents()
        self._set_dimension_hints()

    def _set_dimension_hints(self):
        # Maximum width:
        vwidth = self.verticalHeader().width()
        hwidth = self.horizontalHeader().length()
        swidth = self.style().pixelMetric(QStyle.PixelMetric.PM_ScrollBarExtent)
        fwidth = self.frameWidth() * 2
        current_width = vwidth + hwidth + swidth + fwidth + 1
        # Minimum and maximum height:
        vheight = 0
        for i in range(self.model().rowCount()):
            vheight += self.rowHeight(i)
        hheight = self.horizontalHeader().height()
        sheight = swidth
        fheight = fwidth
        current_height = vheight + hheight + sheight + fheight
        # Check whether the scrollbars are visible:
        if current_height < self.maximum_height:
            # vertical scrollbar not visible
            current_width -= swidth
        if current_width < self.maximum_width:
            current_height -= swidth
        # Set the table dimensions
        if self.maximum_width < current_width:
            self.setMinimumWidth(self.maximum_width)
            self.setMaximumWidth(self.maximum_width)
        else:
            self.setMinimumWidth(current_width)
            self.setMaximumWidth(current_width)
        if self.maximum_height < current_height:
            self.setMinimumHeight(self.maximum_height)
            self.setMaximumHeight(self.maximum_height)
        else:
            self.setMinimumHeight(current_height)
            self.setMaximumHeight(current_height)


class CustomComboBox(QComboBox):
    """QComboBox that allows enabling / disabling items."""

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setModel(_CustomComboBoxModel(parent=self))

    def set_items(self, items):
        self.model().set_items(items)

    def set_item_enabled(self, index, enabled):
        self.model().set_item_enabled(index, enabled)


class _CustomComboBoxModel(QAbstractListModel):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.items = []
        self.enabled = []

    def set_items(self, items):
        self.items = items
        self.enabled = [True] * len(items)
        self.beginResetModel()
        self.endResetModel()

    def set_item_enabled(self, index, enabled):
        self.enabled[index] = enabled
        idx = self.createIndex(index, 0)
        self.dataChanged.emit(idx, idx)

    def rowCount(self, parent=QModelIndex()):
        return len(self.items)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            return self.items[index.row()]

    def flags(self, index):
        if self.enabled[index.row()]:
            return Qt.ItemFlag(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable)
        else:
            return Qt.ItemFlag()



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2018 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import sys
import random
import logging

import numpy as np
import cv2

from .. import sessiondb
from .. import detection
from .. import images
from ..ocr import sample


class LabeledDigit:
    def __init__(self, digit, image_file, corners, identifier=None):
        self.digit = digit
        self.image_file = image_file
        self.corners = corners
        if identifier is not None:
            self.identifier = identifier
        else:
            self.identifier = random.randint(0, 1000000000000)

    def __str__(self):
        data = [self.image_file, str(self.digit)]
        data.extend(str(n) for n in self.corners.reshape(8).tolist())
        return "\t".join(data)

    def crop(self):
        original = images.load_image(self.image_file)
        pre_processed = np.asarray(detection.pre_process(original)[:, :])
        samp = sample.DigitSampleFromCam(self.corners, pre_processed)
        cropped = samp.crop()
        total = cropped.image.shape[0] * cropped.image.shape[1]
        active = sum(sum(cropped.image > 0))
        if active / total >= 0.01:
            cropped_file_path = "digit-{0}-{1}.png".format(self.digit, self.identifier)
            cv2.imwrite(cropped_file_path, cropped.image)
            cropped_digit = LabeledDigit(
                self.digit,
                cropped_file_path,
                cropped.corners,
                identifier=self.identifier,
            )
        else:
            cropped_digit = None
        return cropped_digit


def process_session(labeled_digits, session_path):
    session = sessiondb.SessionDB(session_path)
    for exam in session.exams_iterator():
        image_file = session.get_raw_capture_path(exam["exam_id"])
        cells = session._read_id_cells(exam["exam_id"])
        if exam["student_id"] and len(exam["student_id"]) == len(cells):
            for digit, cell in zip(exam["student_id"], cells):
                corners = np.array([cell.plu, cell.pru, cell.pld, cell.prd])
                labeled_digit = LabeledDigit(int(digit), image_file, corners)
                cropped_digit = labeled_digit.crop()
                if cropped_digit is not None:
                    labeled_digits[int(digit)].append(cropped_digit)
    session.close()


def dump_digit_list(labeled_digits):
    with open("digits.txt", "a") as f:
        for digit in range(10):
            for labeled_digit in labeled_digits[digit]:
                print(str(labeled_digit), file=f)


def _initialize_digits_dict():
    labeled_digits = {}
    for i in range(10):
        labeled_digits[i] = []
    return labeled_digits


def main():
    logging.basicConfig(level=logging.INFO)
    labeled_digits = _initialize_digits_dict()
    for session_path in sys.argv[1:]:
        logging.info("Processing session {}".format(session_path))
        process_session(labeled_digits, session_path)
    dump_digit_list(labeled_digits)


if __name__ == "__main__":
    main()



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#

import unittest
import os.path
import tempfile

import eyegrade.sessiondb as sessiondb
import eyegrade.exams as exams
import eyegrade.students as students


class TestSessionDB(unittest.TestCase):
    def setUp(self):
        self.students = [
            students.Student("101010101", "Donald Duck", "", "", ""),
            students.Student("202020202", "Marty McFly", "", "", ""),
            students.Student("313131313", "Peter Pan", "", "", ""),
        ]
        self.more_students = [
            students.Student("909090909", "Donkey Kong", "", "", ""),
            students.Student("818181818", "Pinoccio", "", "", ""),
            students.Student("555555555", "Gump", "", "", ""),
        ]

    def _get_test_file_path(self, filename):
        dirname = os.path.dirname(os.path.abspath(__file__))
        return os.path.join(dirname, filename)

    def test_exam_data(self):
        exam_config = exams.ExamConfig(filename=self._get_test_file_path("test.eye"))
        exam_config.capture_pattern = "exam-{student-id}-{seq-number}.png"
        listing = students.GroupListing(students.StudentGroup(1, "G"), [])
        listing.add_students(self.students)
        listings = students.StudentListings()
        listings.add_listing(listing)
        with tempfile.TemporaryDirectory() as dir_name:
            session_dir = os.path.join(dir_name, "test_session")
            sessiondb.create_session_directory(session_dir, exam_config, listings)
            session = sessiondb.SessionDB(session_dir)
            # Loading a session always sets variations to a list of zeroes
            # instead of the default empty dictionary.
            # Comparison would fail:
            session.exam_config.question_variations = {}
            self.assertEqual(session.exam_config, exam_config)
            session.close()

    def test_exam_data_model_variations(self):
        exam_config = exams.ExamConfig(
            filename=self._get_test_file_path("test-model-variations.eye")
        )
        exam_config.capture_pattern = "exam-{student-id}-{seq-number}.png"
        listing = students.GroupListing(students.StudentGroup(1, "G"), [])
        listing.add_students(self.students)
        listings = students.StudentListings()
        listings.add_listing(listing)
        with tempfile.TemporaryDirectory() as dir_name:
            session_dir = os.path.join(dir_name, "test_session")
            sessiondb.create_session_directory(session_dir, exam_config, listings)
            session = sessiondb.SessionDB(session_dir)
            # Loading a session always sets variations to a list of zeroes
            # instead of the default empty dictionary.
            # Comparison would fail:
            session.exam_config.question_variations = {}
            self.assertEqual(session.exam_config, exam_config)
            session.close()

    def test_exam_data_weights(self):
        exam_config = exams.ExamConfig(
            filename=self._get_test_file_path("test-weights.eye")
        )
        exam_config.capture_pattern = "exam-{student-id}-{seq-number}.png"
        listing = students.GroupListing(students.StudentGroup(1, "G"), [])
        listing.add_students(self.students)
        listings = students.StudentListings()
        listings.add_listing(listing)
        with tempfile.TemporaryDirectory() as dir_name:
            session_dir = os.path.join(dir_name, "test_session")
            sessiondb.create_session_directory(session_dir, exam_config, listings)
            session = sessiondb.SessionDB(session_dir)
            # Loading a session always sets variations to a list of zeroes
            # instead of the default empty dictionary.
            # Comparison would fail:
            session.exam_config.question_variations = {}
            self.assertEqual(session.exam_config, exam_config)
            session.close()

    def test_student_list(self):
        exam_config = exams.ExamConfig(filename=self._get_test_file_path("test.eye"))
        exam_config.capture_pattern = "exam-{student-id}-{seq-number}.png"
        listing = students.GroupListing(students.StudentGroup(1, "G"), [])
        listing.add_students(self.students)
        listings = students.StudentListings()
        listings.add_listing(listing)
        with tempfile.TemporaryDirectory() as dir_name:
            session_dir = os.path.join(dir_name, "test_session")
            sessiondb.create_session_directory(session_dir, exam_config, listings)
            session = sessiondb.SessionDB(session_dir)
            session.student_listings[0].add_students(self.more_students)
            session.close()
            session = sessiondb.SessionDB(session_dir)
            student_list = list(session.student_listings[0])
            self.assertEqual(
                len(student_list), len(self.students) + len(self.more_students)
            )
            for student_1, student_2 in zip(self.students, student_list):
                self.assertEqual(student_1.name, student_2.name)
                self.assertEqual(student_1.student_id, student_2.student_id)
            for student_1, student_2 in zip(
                self.more_students, student_list[len(self.students) :]
            ):
                self.assertEqual(student_1.name, student_2.name)
                self.assertEqual(student_1.student_id, student_2.student_id)
            session.close()



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Rodrigo Arguello, Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import json
import os

import cv2
import numpy as np

from . import preprocessing
from .. import utils

DEFAULT_DIG_CLASS_FILE = "digit_classifier.dat.gz"
DEFAULT_DIG_META_FILE = "digit_classifier_metadata.txt"
DEFAULT_CROSS_CLASS_FILE = "cross_classifier.dat.gz"
DEFAULT_CROSS_META_FILE = "cross_classifier_metadata.json"
DEFAULT_DIR = "svm"


class SVMClassifier:
    def __init__(self, num_classes, features_extractor, load_from_file=None):
        self.num_classes = num_classes
        self.features_extractor = features_extractor
        if not load_from_file:
            self.svm = cv2.ml.SVM_create()
        else:
            self.svm = cv2.ml.SVM_load(SVMClassifier.resource(load_from_file))

    @property
    def features_len(self):
        return self.features_extractor.features_len

    def train(self, samples, params=None):
        features = np.ndarray(shape=(len(samples), self.features_len), dtype="float32")
        labels = np.ndarray(shape=(len(samples), 1), dtype="int32")
        for i, sample in enumerate(samples):
            features[i, :] = self.features_extractor.extract(sample)
            labels[i] = sample.label
        self.svm.trainAuto(features, cv2.ml.ROW_SAMPLE, labels)

    def classify(self, sample):
        features = np.ndarray(shape=(1, self.features_len), dtype="float32")
        features[0, :] = self.features_extractor.extract(sample)
        retval, prediction = self.svm.predict(features)
        return int(prediction[0, 0])

    def reset(self):
        self.svm = cv2.ml.SVM_create()

    def save(self, filename):
        self.svm.save(filename)

    @staticmethod
    def resource(filename):
        return utils.resource_path(os.path.join(DEFAULT_DIR, filename))


class SVMDigitClassifier(SVMClassifier):
    def __init__(
        self, features_extractor, load_from_file=None, confusion_matrix_from_file=None
    ):
        super().__init__(10, features_extractor, load_from_file=load_from_file)
        self.confusion_matrix = self._load_confusion_matrix(confusion_matrix_from_file)

    def classify_digit(self, sample):
        digit = self.classify(sample)
        weights = self.confusion_matrix[:, digit]
        return (digit, weights)

    @staticmethod
    def _load_confusion_matrix(filename):
        if filename:
            with open(SVMClassifier.resource(filename)) as f:
                metadata = json.load(f)
                matrix = np.array(metadata["confusion_matrix"], dtype=float)
        else:
            matrix = np.diag(np.ones(10, dtype=float))
        return matrix


class DefaultDigitClassifier(SVMDigitClassifier):
    def __init__(
        self,
        load_from_file=DEFAULT_DIG_CLASS_FILE,
        confusion_matrix_from_file=DEFAULT_DIG_META_FILE,
    ):
        super().__init__(
            preprocessing.FeatureExtractor(),
            load_from_file=load_from_file,
            confusion_matrix_from_file=confusion_matrix_from_file,
        )

    def train(self, samples, params=None):
        super().train(samples, dict(C=3.16227766, gamma=0.01))


class SVMCrossesClassifier(SVMClassifier):
    def __init__(self, features_extractor, load_from_file=None):
        super().__init__(2, features_extractor, load_from_file=load_from_file)

    def is_cross(self, sample):
        return self.classify(sample) == 1


class DefaultCrossesClassifier(SVMCrossesClassifier):
    def __init__(self, load_from_file=DEFAULT_CROSS_CLASS_FILE):
        super().__init__(
            preprocessing.CrossesFeatureExtractor(), load_from_file=load_from_file
        )

    def train(self, samples, params=None):
        super().train(samples, dict(C=100, gamma=0.01))



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import fractions
import re
import decimal
import enum

from typing import Union, Optional, Sequence

from . import utils


class AnswerStatus(enum.Enum):
    CORRECT = 1
    INCORRECT = 2
    BLANK = 3
    VOID = 4


class QuestionScores(utils.ComparableMixin):
    """Compute the score of a question."""

    correct_score: Union[int, float, decimal.Decimal, fractions.Fraction]
    incorrect_score: Union[int, float, decimal.Decimal, fractions.Fraction]
    blank_score: Union[int, float, decimal.Decimal, fractions.Fraction]
    weight: Union[int, float, decimal.Decimal, fractions.Fraction]
    _correct_score_internal: Union[int, float, fractions.Fraction]
    _incorrect_score_internal: Union[int, float, fractions.Fraction]
    _blank_score_internal: Union[int, float, fractions.Fraction]
    _weight_internal: Union[int, float, fractions.Fraction]

    def __init__(
        self,
        correct_score: Union[str, int, float, decimal.Decimal, fractions.Fraction],
        incorrect_score: Union[str, int, float, decimal.Decimal, fractions.Fraction],
        blank_score: Union[str, int, float, decimal.Decimal, fractions.Fraction],
        weight: Union[str, int, float, decimal.Decimal, fractions.Fraction] = 1,
    ):
        if isinstance(correct_score, str):
            self.correct_score = self._parse_score(correct_score)
        else:
            self.correct_score = correct_score
        if isinstance(incorrect_score, str):
            self.incorrect_score = self._parse_score(
                incorrect_score, invert_negatives=True
            )
        else:
            self.incorrect_score = incorrect_score
        if isinstance(blank_score, str):
            self.blank_score = self._parse_score(blank_score, invert_negatives=True)
        else:
            self.blank_score = blank_score
        if isinstance(weight, str):
            self.weight = self._parse_weight(weight)
        else:
            self.weight = weight
        self._correct_score_internal = _value_for_computations(self.correct_score)
        self._incorrect_score_internal = _value_for_computations(self.incorrect_score)
        self._blank_score_internal = _value_for_computations(self.blank_score)
        self._weight_internal = _value_for_computations(self.weight)

    def score(self, answer_type: AnswerStatus) -> Union[int, float, fractions.Fraction]:
        if answer_type == AnswerStatus.CORRECT:
            return self._weight_internal * self._correct_score_internal
        elif answer_type == AnswerStatus.INCORRECT:
            return -self._weight_internal * self._incorrect_score_internal
        elif answer_type == AnswerStatus.BLANK:
            return -self._weight_internal * self._blank_score_internal
        elif answer_type == AnswerStatus.VOID:
            return 0
        else:
            raise Exception("Bad answer_type value in QuestionScore")

    def format_all(self) -> str:
        data = (
            self._format_score(self.correct_score),
            self._format_score(self.incorrect_score),
            self._format_score(self.blank_score),
        )
        return ";".join(data)

    def format_weight(self) -> str:
        return self._format_score(self.weight)

    def format_score(self, answer_type: AnswerStatus, signed: bool = False) -> str:
        if answer_type == AnswerStatus.CORRECT:
            return self._format_score(self.correct_score, signed=False)
        elif answer_type == AnswerStatus.INCORRECT:
            return self._format_score(self.incorrect_score, signed=signed)
        elif answer_type == AnswerStatus.BLANK:
            return self._format_score(self.blank_score, signed=signed)
        else:
            raise ValueError("Bad answer_type value in QuestionScore")

    def format_correct_score(self, signed: bool = False):
        return self._format_score(self.correct_score, signed=False)

    def format_incorrect_score(self, signed: bool = False):
        return self._format_score(self.incorrect_score, signed=signed)

    def format_blank_score(self, signed: bool = False):
        return self._format_score(self.blank_score, signed=signed)

    def clone(
        self,
        new_weight: Union[int, float, decimal.Decimal, fractions.Fraction, None] = None,
    ) -> "QuestionScores":
        if new_weight is not None:
            weight = new_weight
        else:
            weight = self.weight
        return QuestionScores(
            self.correct_score, self.incorrect_score, self.blank_score, weight=weight
        )

    def __str__(self) -> str:
        return "({0}) * {1}".format(self.format_all(), self.format_weight())

    def _parse_score(
        self, score_str, invert_negatives=False
    ) -> Union[int, float, decimal.Decimal, fractions.Fraction]:
        score = parse_number(score_str, allow_negatives=invert_negatives)
        if score < 0:
            score = -score
        return score

    def _parse_weight(
        self, score_str: str
    ) -> Union[int, float, decimal.Decimal, fractions.Fraction]:
        score = parse_number(score_str)
        if score < 0:
            raise ValueError("Negative weights are forbidden: {}".format(score_str))
        return score

    def _format_score(
        self,
        score: Union[int, float, decimal.Decimal, fractions.Fraction],
        signed: bool = False,
    ):
        if signed:
            score = -score
        return format_number(score)

    def _cmpkey(self) -> tuple:
        return (self.correct_score, self.incorrect_score, self.blank_score, self.weight)


class Score:
    correct: Optional[int]
    incorrect: Optional[int]
    blank: Optional[int]
    score: Optional[float]
    max_score: Optional[float]
    answer_status: Optional[list[AnswerStatus]]
    answers: Optional[list[int]]
    solutions: Optional[list[set[int]]]
    question_scores: Optional[list[QuestionScores]]

    def __init__(
        self,
        answers: Optional[list[int]],
        solutions: Optional[list[set[int]]],
        question_scores: Optional[list[QuestionScores]],
    ) -> None:
        if answers is not None and solutions and len(answers) != len(solutions):
            raise ValueError("Parameters must have the same length in Score")
        if (
            solutions
            and question_scores is not None
            and len(solutions) != len(question_scores)
        ):
            raise ValueError("Parameters must have the same length in Score")
        self.correct = None
        self.incorrect = None
        self.blank = None
        self.score = None
        self.max_score = None
        self.answer_status = None
        self.answers = answers
        self.solutions = solutions
        self.question_scores = question_scores
        self.update()

    def update(self) -> None:
        if not self.answers or not self.solutions:
            return
        self.correct = 0
        self.incorrect = 0
        self.blank = 0
        self.answer_status = []
        internal_scores: Sequence[Optional[QuestionScores]]
        if self.question_scores is not None:
            internal_scores = self.question_scores
        else:
            internal_scores = [None] * len(self.answers)
        for answer, solution, q in zip(self.answers, self.solutions, internal_scores):
            if q is not None and q.weight == 0:
                self.answer_status.append(AnswerStatus.VOID)
            elif answer == 0:
                self.blank += 1
                self.answer_status.append(AnswerStatus.BLANK)
            elif answer in solution:
                self.correct += 1
                self.answer_status.append(AnswerStatus.CORRECT)
            else:
                self.incorrect += 1
                self.answer_status.append(AnswerStatus.INCORRECT)
        if self.question_scores:
            self.score = float(
                sum(
                    [
                        q.score(status)
                        for q, status in zip(self.question_scores, self.answer_status)
                    ]
                )
            )
            self.max_score = float(
                sum([q.score(AnswerStatus.CORRECT) for q in self.question_scores])
            )
        else:
            self.score = None
            self.max_score = None

    def update_question_scores(self, question_scores: list[QuestionScores]) -> None:
        self.question_scores = question_scores
        self.update()


class AutomaticScore:
    max_score: Union[int, float, decimal.Decimal, fractions.Fraction]
    penalize: bool

    def __init__(
        self,
        max_score: Union[str, int, float, decimal.Decimal, fractions.Fraction],
        penalize: bool,
    ) -> None:
        if isinstance(max_score, str):
            self.max_score = parse_number(max_score)
        else:
            self.max_score = max_score
        self.penalize = penalize

    def compute(self, num_questions: int, num_choices: Optional[int]) -> QuestionScores:
        correct_score = self.max_score / num_questions
        if self.penalize:
            if num_choices is not None:
                incorrect_score = self.max_score / (num_choices - 1) / num_questions
            else:
                raise ValueError(
                    "num_choices must be provided when penalizing (null received)"
                )
        else:
            incorrect_score = 0
        return QuestionScores(correct_score, incorrect_score, 0)


def format_number(
    number: Union[int, float, decimal.Decimal, fractions.Fraction, None],
    short: bool = False,
    no_fraction: bool = False,
) -> Union[str, None]:
    if number is None:
        return None
    elif no_fraction and type(number) == fractions.Fraction:
        if number.denominator != 1:
            number = float(number)
    if type(number) == fractions.Fraction:
        if number.denominator != 1:
            return "{0}/{1}".format(number.numerator, number.denominator)
        else:
            return str(number.numerator)
    elif type(number) == float:
        if short:
            return "{0:.2f}".format(number)
        else:
            return "{0:.16f}".format(number)
    else:
        # e.g. decimal.Decimal or int
        return str(number)


# A score is a decimal.Decimal, float or a fraction, e.g.: '0.8' or '4/5'
_re_number = re.compile(r"^(-?)\s*((\d+(\.\d+)?)|((\d+)\s*\/\s*(\d+)))\s*$")
# score_re = re.compile(r'^(-?)((\d*(\.\d+))|((\d+)(\/(\d+))?))$')


def parse_number(
    score_str: str, force_float: bool = False, allow_negatives: bool = False
) -> Union[int, float, decimal.Decimal, fractions.Fraction]:
    value: Union[int, float, decimal.Decimal, fractions.Fraction]
    match = _re_number.match(score_str)
    if match is None:
        raise ValueError("Syntax error in score: " + score_str)
    groups = match.groups()
    sign = -1 if groups[0] else 1
    if sign == -1 and not allow_negatives:
        raise ValueError("The number cannot be negative: " + score_str)
    if groups[2] is not None:
        if groups[3] is not None:
            if force_float:
                value = sign * float(groups[2])
            else:
                value = sign * decimal.Decimal(groups[2])
            numerator = None
        else:
            numerator = int(groups[2])
            denominator = 1
    else:
        numerator = int(groups[5])
        denominator = int(groups[6])
    if numerator is not None:
        value = fractions.Fraction(sign * numerator, denominator)
        if force_float:
            value = float(value)
    return value


def _value_for_computations(
    value: Union[int, float, decimal.Decimal, fractions.Fraction],
) -> Union[int, float, fractions.Fraction]:
    if isinstance(value, decimal.Decimal):
        return fractions.Fraction(value)
    else:
        return value



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import configparser
import csv
import os
import locale
import codecs
import sys
import re
import contextlib

from typing import Dict


program_name = "eyegrade"
web_location = "https://www.eyegrade.org/"
source_location = "https://github.com/jfisteus/eyegrade"
help_location = "https://www.eyegrade.org/doc/user-manual/"
version = "1.0rc1"

re_model_letter = re.compile("[0a-zA-Z]")

csv_tabs_dialect = "tabs"
csv.register_dialect(csv_tabs_dialect, delimiter=str("\t"))

results_file_keys = [
    "seq-num",
    "student-id",
    "model",
    "good",
    "bad",
    "score",
    "answers",
]

default_capture_pattern = "exam-{student-id}-{seq-number}.png"

# The data_dir variable will be intially None.  The functions in this
# module that depend on the data directory will initialize it if
# needed.
data_dir = None


class ComparableMixin:
    """For implementing comparable classes.

    As seen on http://regebro.wordpress.com/2010/12/13/
                      python-implementing-rich-comparison-the-correct-way/

    """

    def _compare(self, other, method):
        try:
            return method(self._cmpkey(), other._cmpkey())
        except (AttributeError, TypeError):
            # _cmpkey not implemented, or return different type,
            # so I can't compare with "other".
            return NotImplemented

    def __lt__(self, other):
        return self._compare(other, lambda s, o: s < o)

    def __le__(self, other):
        return self._compare(other, lambda s, o: s <= o)

    def __eq__(self, other):
        return self._compare(other, lambda s, o: s == o)

    def __ge__(self, other):
        return self._compare(other, lambda s, o: s >= o)

    def __gt__(self, other):
        return self._compare(other, lambda s, o: s > o)

    def __ne__(self, other):
        return self._compare(other, lambda s, o: s != o)


def user_home():
    return os.path.expanduser("~/")


def _read_config():
    """Reads the general config file and returns the resulting config object.

    Other modules can get the config object by accessing the
    utils.config variable.

    """
    conf = {
        "camera-dev": "0",
        "save-filename-pattern": default_capture_pattern,
        "csv-dialect": "tabs",
        "default-charset": "utf8",  # special value: 'system-default'
    }
    parser = configparser.ConfigParser()
    home = user_home()
    try:
        parser.read(
            [
                os.path.join(home, ".eyegrade.cfg"),
                os.path.join(home, ".camgrade.cfg"),
                resource_path("default.cfg"),
            ]
        )
    except EyegradeException:
        parser.read(
            [os.path.join(home, ".eyegrade.cfg"), os.path.join(home, ".camgrade.cfg")]
        )
    if "default" in parser.sections():
        for option in parser.options("default"):
            conf[option] = parser.get("default", option)
    if not conf["csv-dialect"] in csv.list_dialects():
        conf["csv-dialect"] = "tabs"
    if "error-logging" in conf and conf["error-logging"] == "yes":
        conf["error-logging"] = True
    else:
        conf["error-logging"] = False
    conf["camera-dev"] = int(conf["camera-dev"])
    if conf["default-charset"] == "system-default":
        conf["default-charset"] = locale.getpreferredencoding()
    if "gui-styles" in conf:
        conf["gui-styles"] = tuple(v.strip() for v in conf["gui-styles"].split(","))
    else:
        conf["gui-styles"] = None
    return conf


class EyegradeException(Exception):
    """An Eyegrade-specific exception.

    In addition to what a normal exception would do, it encapsulates
    user-friendly messages for some common causes of error due to
    the user.

    """

    _error_messages: Dict[str, str] = {}
    _short_messages: Dict[str, str] = {}

    def __init__(self, message, key=None, format_params=None):
        """Creates a new exception.

        If `key` is in `_error_messages`, a prettier version of the
        exception will be shown to the user, with the explanation appended
        to the end of what you provide in `message`.

        """
        self.key = key
        if (
            key in EyegradeException._error_messages
            or key in EyegradeException._short_messages
        ):
            parts = []
            if message:
                parts.append(message)
            elif key in EyegradeException._short_messages:
                short_msg = EyegradeException._short_messages[key]
                if not format_params:
                    parts.append(short_msg)
                else:
                    parts.append(short_msg.format(*format_params))
            if key in EyegradeException._error_messages:
                parts.append("\n\n")
                parts.append(EyegradeException._error_messages[key])
            parts.append("\n")
            self.full_message = "\n".join(parts)
            super().__init__(self.full_message)
        else:
            self.full_message = None
            super().__init__(message)

    def __str__(self):
        """Prints the exception.

        A user-friendly message, without the stack trace, is shown when such
        user-friendly message is available.

        """
        if self.full_message is not None:
            return self.full_message
        else:
            return super().__str__()

    @staticmethod
    def register_error(key, detailed_message="", short_message=""):
        """Registers a new error message associated to a key.

        `key` is just a string used to identify this error message,
        that must be passed when creating exception
        objects. `detailed_message` is a (possibly long and with end
        of line characters inside) explanation of the
        error. `short_message` is a one line error message to be used
        only when a blank message is passed when creating the
        exception.

        Being this method static, messages added through it will be
        shared for all the instances of the exception.

        """
        if (
            key not in EyegradeException._error_messages
            and key not in EyegradeException._short_messages
        ):
            if detailed_message:
                EyegradeException._error_messages[key] = detailed_message
            if short_message:
                EyegradeException._short_messages[key] = short_message
        else:
            raise EyegradeException("Duplicate error key in register_error")


EyegradeException.register_error(
    "bad_dimensions",
    "Dimensions must be specified as a ';' separated list of tables.\n"
    "For each table, specify the number of choices + comma + the number of\n"
    "questions in that table. For example, '4,10;4,9' configures two\n"
    "tables, the left-most with 9 questions and 4 choices per question,\n"
    "and the right-most with 10 questions and the same number of choices."
    "Bad dimensions value.",
)

EyegradeException.register_error(
    "same_num_choices",
    "By now, Eyegrade needs you to use the same number of choices in\n"
    "all the questions of the exam.",
    "There are questions with a different number of choices",
)

_student_list_message = (
    "The file is expected to contain one line per student.\n"
    "Each line can contain one or more TAB-separated columns.\n"
    "The first column must be the student id (a number).\n"
    "The second column, if present, is interpreted as the student name.\n"
    "The rest of the columns are ignored."
)

EyegradeException.register_error(
    "load_image", "The image cannot be loaded (perhaps a wrong path?)."
)


def guess_data_dir():
    path = os.path.split(os.path.realpath(__file__))[0]
    # An alternative path to try for pyinstaller's packages:
    path_alt = os.path.split(path)[0]
    paths_to_try = [
        os.path.join(path, "data"),
        os.path.join(path, "..", "data"),
        os.path.join(path, "..", "..", "data"),
        os.path.join(path, "..", "..", "..", "data"),
        os.path.join(path_alt, "data"),
    ]
    for p in paths_to_try:
        if os.path.isdir(p):
            return os.path.abspath(p)
    raise EyegradeException("Data path not found!")


def init_data_dir():
    global data_dir
    data_dir = guess_data_dir()


def locale_dir():
    if data_dir is None:
        init_data_dir()
    return os.path.join(data_dir, "locale")


def qt_translations_dir():
    if data_dir is None:
        init_data_dir()
    return os.path.join(data_dir, "qt-translations")


def resource_path(file_name):
    if data_dir is None:
        init_data_dir()
    return os.path.join(data_dir, file_name)


# The global configuration object:
config = _read_config()


def check_model_letter(model, allow_question_mark=False):
    """Checks if a model letter is correct.

    The special value '?' is considered valid only if the parameter
    `allow_question_mark` is set.

    """
    if re_model_letter.match(model):
        return model.upper()
    elif allow_question_mark and model == "?":
        return "?"
    else:
        raise Exception("Incorrect model letter: " + model)


def _permute_answers(answers, permutation):
    assert len(answers) == len(permutation)
    permutted = [0] * len(answers)
    for i, option in enumerate(answers):
        if option == 0 or option == -1:
            resolved_option = option
        else:
            resolved_option = permutation[i][1][option - 1]
        permutted[permutation[i][0] - 1] = resolved_option
    return permutted


def encode_model(model, num_tables, num_answers):
    """Given the letter of the model, returns the infobits pattern.

    It is formatted as an array of booleans string where the pos. 0
    is the one that goes in the column of the table at the left.
    The length of the string is 'num_tables' * 'num_answers', where
    'num_tables' is the number of answer tables and 'num_tables'
    the number of answers per question. The 'model' must be a
    capital ASCII letter.

    """
    if len(model) != 1 or model < "A" or model > "Z":
        raise Exception("Incorrect model letter")
    if model > "H":
        raise Exception("Model is currently limited to A - H")
    model_num = ord(model) - 65
    num_bits = num_tables * num_answers
    if model_num >= 2**num_bits:
        raise Exception("Model number too big given the number of answers")
    seed = _int_to_bin(model_num, 3, True)
    seed[2] = not seed[2]
    seed.append(seed[0] ^ seed[1] ^ seed[2])
    seed[2] = not seed[2]
    bit_list = seed * (1 + (num_bits - 1) // 4)
    return bit_list[: num_tables * num_answers]


def decode_model(bit_list, accept_model_0=False):
    """Given the bits that encode the model, returns the associated letter.

    It decoding/checksum fails, None is returned. The list of bits must
    be a list of boolean variables.

    The special model 0 is not valid unless `accept_model_0` is set.

    """
    # x3 = x0 ^ x1 ^ not x2; x0-x3 == x4-x7 == x8-x11 == ...
    valid = False
    if len(bit_list) == 2 or len(bit_list) == 3:
        valid = True
    elif len(bit_list) >= 4:
        if bit_list[3] == bit_list[0] ^ bit_list[1] ^ (not bit_list[2]):
            valid = True
            for i in range(4, len(bit_list)):
                if bit_list[i] != bit_list[i - 4]:
                    valid = False
                    break
    if valid:
        return chr(
            65 + (int(bit_list[0]) | int(bit_list[1]) << 1 | int(bit_list[2]) << 2)
        )
    elif accept_model_0 and max(bit_list) is False:
        return "0"
    else:
        return None


def _int_to_bin(n, num_digits, reverse=False):
    """Returns the binary representation of a number as a list of booleans.

    If the number of digits is less than 'num_digits', it is
    completed with False in the most-significative side. If
    'reverse' is True returns the least significative bit in the
    first position of the string.

    There is a bin() function in python >= 2.6, but by now we want
    the program to be compatible with 2.5. Anyway, the behaviour of
    that function is different.

    """
    binary = []
    while n > 0:
        n, remainder = divmod(n, 2)
        binary.append(bool(remainder))
    if len(binary) < num_digits:
        binary.extend([False] * (num_digits - len(binary)))
    if reverse:
        return binary
    return binary[::-1]


def read_file(file_name):
    """Return the contents of a file as a Unicode string using terminal locale."""
    file_ = codecs.open(file_name, "r", config["default-charset"])
    data = file_.read()
    file_.close()
    return data


def write_file(file_name, unicode_text):
    """Writes a Unicode string in a file using terminal locale."""
    file_ = codecs.open(file_name, "w", config["default-charset"])
    file_.write(unicode_text)
    file_.close()


def write_to_stdout(unicode_text):
    """Writes a Unicode string in sys.stdout using terminal locale."""
    sys.stdout.write(unicode_text)


def increment_list(list_):
    """Adds one to every element in a list of integers. Returns a new list."""
    return [n + 1 for n in list_]


def parse_dimensions(text, check_equal_num_choices=False):
    dimensions = []
    num_options = []
    boxes = text.split(";")
    for box in boxes:
        dims = box.split(",")
        try:
            data = (int(dims[0]), int(dims[1]))
        except ValueError:
            raise EyegradeException(
                "Incorrect number in exam dimensions", "bad_dimensions"
            )
        if data[0] <= 0 or data[1] <= 0:
            raise EyegradeException(
                "Incorrect number in exam dimensions", "bad_dimensions"
            )
        dimensions.append(data)
        num_options.extend([data[0]] * data[1])
    if len(dimensions) == 0:
        raise EyegradeException("Dimensions are empty", "bad_dimensions")
    if check_equal_num_choices:
        for i in range(1, len(dimensions)):
            if dimensions[i][0] != dimensions[0][0]:
                raise EyegradeException("", "same_num_choices")
    return dimensions, num_options


# Regular expressions used for capture filename patterns
regexp_id = re.compile(r"\{student-id\}")
regexp_seqnum = re.compile(r"\{seq-number\}")


def capture_name(filename_pattern, exam_id, student):
    if student is not None:
        sid = student.student_id
    else:
        sid = "noid"
    filename = regexp_seqnum.sub(str(exam_id), filename_pattern)
    filename = regexp_id.sub(sid, filename)
    return filename


@contextlib.contextmanager
def change_dir(directory):
    prev_directory = os.getcwd()
    if directory:
        os.chdir(directory)
    yield
    if directory:
        os.chdir(prev_directory)



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import re

from typing import Dict, List, Tuple

from .. import utils
from . import questions

_PARAM_REPLACE_RE = re.compile("{{([^{}]+)}}")

EyegradeException = utils.EyegradeException

EyegradeException.register_error(
    "undefined_parameter",
    "A parameter declaration is missing in a parametric question variation.",
)


class ParametricQuestion(questions.Question):
    """Question with variations defined through a common pattern and varying parameter sets."""

    question_pattern: questions.QuestionVariation
    parameter_sets: List["ParameterSet"]

    def __init__(self, question_pattern: questions.QuestionVariation) -> None:
        super().__init__()
        self.question_pattern = question_pattern
        self.parameter_sets = []

    def add_parameter_set(self, parameter_set: "ParameterSet"):
        self.parameter_sets.append(parameter_set)
        super().add_variation(parameter_set.apply_to(self.question_pattern))


class ParametricGroupCommonComponent(questions.GroupCommonComponent):
    """Group common component with parametric variations."""

    component_pattern: questions.QuestionComponent
    parameter_sets: List["ParameterSet"]

    def __init__(self, component_pattern: questions.QuestionComponent) -> None:
        super().__init__()
        self.component_pattern = component_pattern
        self.parameter_sets = []

    def add_parameter_set(self, parameter_set: "ParameterSet"):
        self.parameter_sets.append(parameter_set)
        super().add_variation(
            parameter_set.apply_to_question_component(self.component_pattern)
        )


class ParameterSet:
    """Container for a set of name/value parameters."""

    parameters: Dict[str, str]

    def __init__(self):
        self.parameters = {}

    def add_parameter(self, name: str, value: str) -> None:
        self.parameters[name] = value

    def apply_to(
        self, question_pattern: questions.QuestionVariation
    ) -> questions.QuestionVariation:
        text = self.apply_to_question_component(question_pattern.text)
        correct_choices = [
            self.apply_to_question_component(choice)
            for choice in question_pattern.correct_choices
        ]
        incorrect_choices = [
            self.apply_to_question_component(choice)
            for choice in question_pattern.incorrect_choices
        ]
        return questions.QuestionVariation(
            text,
            correct_choices,
            incorrect_choices,
            question_pattern.fix_first,
            question_pattern.fix_last,
        )

    def apply_to_question_component(
        self, component: questions.QuestionComponent
    ) -> questions.QuestionComponent:
        replaced = questions.QuestionComponent(component.in_choice)
        replaced.annex_width = component.annex_width
        replaced.annex_pos = component.annex_pos
        text = component.text
        if text is None:
            replaced.text = None
        elif isinstance(text, str):
            replaced.text = self._apply_to_text(text)
        else:
            replaced.text = [self._apply_to_text_part(part) for part in text]
        if component.code is not None:
            replaced.code = self._apply_to_text(component.code)
        if component.figure is not None:
            replaced.figure = self._apply_to_text(component.figure)
        return replaced

    def _apply_to_text_part(self, part: Tuple[str, str]) -> Tuple[str, str]:
        if part[0] == "text":
            return ("text", self._apply_to_text(part[1]))
        return part

    def _apply_to_text(self, text: str) -> str:
        replaced: List[str]
        parts = _PARAM_REPLACE_RE.split(text)
        # Replacement keys are at odd positions of parts
        replaced = len(parts) * [""]
        replaced[::2] = parts[::2]
        try:
            replaced[1::2] = [self.parameters[param_name] for param_name in parts[1::2]]
        except KeyError as exception:
            raise EyegradeException(
                "Parameter: {}".format(exception.args[0]), key="undefined_parameter"
            )
        return "".join(replaced)
